import configparser
import logging
import os
import re
import subprocess
from pathlib import Path

__all__ = (
    "FLATPAK_BWRAP_COMPATIBLE_VERSION", "FLATPAK_INFO_PATH",
    "is_flatpak_sandbox", "get_running_flatpak_version",
    "get_inaccessible_paths"
)

logger = logging.getLogger("protontricks")

# Flatpak minimum version required to enable bwrap. In other words, the first
# Flatpak version with the necessary support for sub-sandboxes.
FLATPAK_BWRAP_COMPATIBLE_VERSION = (1, 12, 1)

FLATPAK_INFO_PATH = "/.flatpak-info"


def is_flatpak_sandbox():
    """
    Check if we're running inside a Flatpak sandbox
    """
    return bool(get_running_flatpak_version())


def _get_flatpak_config():
    config = configparser.ConfigParser()

    try:
        config.read_string(Path(FLATPAK_INFO_PATH).read_text(encoding="utf-8"))
    except FileNotFoundError:
        return None

    return config


_XDG_PERMISSIONS = {
    "xdg-desktop": "DESKTOP",
    "xdg-documents": "DOCUMENTS",
    "xdg-download": "DOWNLOAD",
    "xdg-music": "MUSIC",
    "xdg-pictures": "PICTURES",
    "xdg-public-share": "PUBLICSHARE",
    "xdg-videos": "VIDEOS",
    "xdg-templates": "TEMPLATES",
}


def _get_xdg_user_dir(permission):
    """
    Get the XDG user directory corresponding to the given "xdg-" prefixed
    Flatpak permission and retrieve its absolute path using the `xdg-user-dir`
    command.
    """
    if permission in _XDG_PERMISSIONS:
        # This will only be called in a Flatpak environment, and we can assume
        # 'xdg-user-dir' always exists in that environment.
        path = subprocess.check_output(
            ["xdg-user-dir", _XDG_PERMISSIONS[permission]]
        )
        path = path.strip()
        path = os.fsdecode(path)
        logger.debug("XDG path for %s is %s", permission, path)
        return Path(path)

    return None


def get_running_flatpak_version():
    """
    Get the running Flatpak version if running inside a Flatpak sandbox,
    or None if Flatpak sandbox isn't active
    """
    config = _get_flatpak_config()

    if config is None:
        return None

    # If this fails it's because the Flatpak version is older than 0.6.10.
    # Since Steam Flatpak requires at least 1.0.0, we can fail here instead
    # of continuing on. It's also extremely unlikely, since even older distros
    # like CentOS 7 ship Flatpak releases newer than 1.0.0.
    version = config["Instance"]["flatpak-version"]

    # Remove non-numeric characters just in case (eg. if a suffix like '-pre'
    # is used).
    version = "".join([ch for ch in version if ch in ("0123456789.")])

    # Convert version number into a tuple
    version = tuple([int(part) for part in version.split(".")])
    return version


def get_inaccessible_paths(paths):
    """
    Check which given paths are inaccessible under Protontricks.

    Inaccessible paths are returned as a list. This has no effect in
    non-Flatpak environments, where an empty list is always returned.
    """
    def _path_is_relative_to(a, b):
        try:
            a.relative_to(b)
            return True
        except ValueError:
            return False

    def _map_path(path):
        if path == "":
            return None

        if path.startswith("xdg-data/"):
            return (
                Path("~/.local/share").expanduser()
                / path.split("xdg-data/")[1]
            )

        if path.startswith("xdg-"):
            path_ = _get_xdg_user_dir(path)

            if path_:
                return path_

        if path == "home":
            return Path.home()

        if path.startswith("/"):
            return Path(path).resolve()

        if path.startswith("~"):
            return Path(path).expanduser()

        logger.warning(
            "Unknown Flatpak file system permission '%s', ignoring.",
            path
        )
        return None

    if not is_flatpak_sandbox():
        return []

    config = _get_flatpak_config()

    try:
        mounted_paths = \
            re.split(r'(?<!\\);', config["Context"]["filesystems"])
    except KeyError:
        logger.warning("Could not find mounted Flatpak filesystems")
        return []

    if "host" in mounted_paths:
        # If 'host' is enabled, Flatpak has full file system access,
        # aside from some Flatpak specific paths that are not relevant
        # for Protontricks or Steam usage.
        return []

    paths = [Path(path).resolve() for path in paths]

    # Resolve the mounted filesystems
    mounted_paths = [_map_path(path) for path in mounted_paths]
    mounted_paths = list(filter(bool, mounted_paths))

    return [
        Path(path) for path in paths
        if not any(
            True for mounted_path in mounted_paths
            if _path_is_relative_to(path, mounted_path)
        )
    ]



import functools
import itertools
import json
import logging
import os
import shlex
import shutil
import sys
from pathlib import Path
from subprocess import PIPE, CalledProcessError, run

import pkg_resources
from PIL import Image

from .config import get_config
from .flatpak import get_inaccessible_paths
from .util import get_cache_dir
from .steam import SNAP_STEAM_DIRS

APP_ICON_SIZE = (32, 32)


__all__ = (
    "LocaleError", "get_gui_provider", "select_steam_app_with_gui",
    "select_steam_installation", "show_text_dialog", "prompt_filesystem_access"
)

logger = logging.getLogger("protontricks")


class LocaleError(Exception):
    pass


@functools.lru_cache(maxsize=1)
def get_gui_provider():
    """
    Get the GUI provider used to display dialogs.
    Returns either 'yad' or 'zenity', preferring 'yad' if both exist.
    """
    try:
        candidates = ["yad", "zenity"]
        # Allow overriding the GUI provider using an envvar
        if os.environ.get("PROTONTRICKS_GUI", "").lower() in candidates:
            candidates.insert(0, os.environ["PROTONTRICKS_GUI"].lower())

        cmd = next(cmd for cmd in candidates if shutil.which(cmd))
        logger.info("Using '%s' as GUI provider", cmd)

        return cmd
    except StopIteration as exc:
        raise FileNotFoundError(
            "'yad' or 'zenity' was not found. Either executable is required "
            "for Protontricks GUI."
        ) from exc


def _get_appid2icon(steam_apps):
    """
    Get icons for Steam apps to show in the app selection dialog.
    Return a {appid: icon_path} dict.
    """
    placeholder_path = Path(
        pkg_resources.resource_filename(
            "protontricks", "data/data/icon_placeholder.png"
        )
    )

    protontricks_icon_dir = get_cache_dir() / "app_icons"
    protontricks_icon_dir.mkdir(exist_ok=True)

    appid2icon = {}

    for app in steam_apps:
        # Use library icon for Steam apps, fallback to placeholder icon
        # for non-Steam shortcuts and missing icons
        icon_cache_path = protontricks_icon_dir / f"{app.appid}.jpg"

        # What path to actually use for the app selector icon
        final_icon_path = placeholder_path

        if app.icon_path:
            # Resize icons that have a non-standard size to ensure they can be
            # displayed consistently in the app selector
            try:
                with Image.open(app.icon_path) as img:
                    # Icon exists, so use the current icon instead of the
                    # default placeholder.
                    final_icon_path = app.icon_path

                    resize_icon = img.size != APP_ICON_SIZE

                    # Resize icons that have a non-standard size to ensure they can
                    # be displayed consistently in the app selector
                    if resize_icon:
                        logger.info(
                            "App icon %s has unusual size, resizing",
                            app.icon_path
                        )
                        resized_img = img.resize(APP_ICON_SIZE).convert("RGB")
                        resized_img.save(icon_cache_path)
                        final_icon_path = icon_cache_path
            except FileNotFoundError:
                # Icon does not exist, the placeholder will be used
                pass
            except Exception:
                # Multitude of reasons can cause image parsing or resizing
                # to fail. Instead of trying to catch everything, log the error
                # and move on.
                logger.warning(
                    "Could not resize %s, ignoring",
                    app.icon_path,
                    exc_info=True
                )

        appid2icon[app.appid] = final_icon_path

    return appid2icon


def _run_gui(args, input_=None, strip_nonascii=False):
    """
    Run YAD/Zenity with the given args.

    If 'strip_nonascii' is True, strip non-ASCII characters to workaround
    environments that can't handle all characters
    """
    if strip_nonascii:
        # Convert to bytes and back to strings while stripping
        # non-ASCII characters
        args = [
            arg.encode("ascii", "ignore").decode("ascii") for arg in args
        ]
        if input_:
            input_ = input_.encode("ascii", "ignore").decode("ascii")

    if input_:
        input_ = input_.encode("utf-8")

    try:
        return run(
            args, input=input_, check=True, stdout=PIPE, stderr=PIPE,
        )
    except CalledProcessError as exc:
        if exc.returncode == 255 and not strip_nonascii:
            # User has weird locale settings. Log a warning and
            # rerun the command while stripping non-ASCII characters.
            logger.warning(
                "Your system locale is incapable of displaying all "
                "characters. Some app names may not show up correctly. "
                "Please use an UTF-8 locale to avoid this warning."
            )
            return _run_gui(args, strip_nonascii=True)

        raise

def show_text_dialog(
        title,
        text,
        window_icon,
        cancel_label=None,
        add_cancel_button=False,
        ok_label=None,
        width=600,
        height=600):
    """
    Show a text dialog to the user

    :returns: True if user clicked OK, False otherwise
    """
    if not ok_label:
        ok_label = "OK"

    if not cancel_label:
        cancel_label = "Cancel"

    def _get_yad_args():
        args = [
            "yad", "--text-info", "--window-icon", window_icon,
            "--title", title, "--width", str(width), "--height", str(height),
            f"--button={ok_label}:0", "--wrap",
            "--margins", "2", "--center"
        ]

        if add_cancel_button:
            args += [f"--button={cancel_label}:1"]

        return args

    def _get_zenity_args():
        args = [
            "zenity", "--text-info", "--window-icon", window_icon,
            "--title", title, "--width", str(width), "--height",
            str(height), "--cancel-label", cancel_label, "--ok-label", ok_label
        ]

        return args

    gui_provider = get_gui_provider()
    if gui_provider == "yad":
        args = _get_yad_args()
    else:
        args = _get_zenity_args()

    process = run(args, input=text.encode("utf-8"), check=False)

    return process.returncode == 0


def select_steam_installation(steam_installations):
    """
    Prompt the user to select a Steam installation if more than one
    installation is available

    Return the selected (steam_path, steam_root) installation, or None
    if the user picked nothing
    """
    def _get_yad_args():
        return [
            "yad", "--list", "--no-headers", "--center",
            "--window-icon", "wine",
            # Disabling markup means we won't have to escape special characters
            "--no-markup",
            "--width", "600", "--height", "400",
            "--text", "Select Steam installation",
            "--title", "Protontricks",
            "--column", "Path"
        ]

    def _get_zenity_args():
        return [
            "zenity", "--list", "--hide-header",
            "--width", "600",
            "--height", "400",
            "--text", "Select Steam installation",
            "--title", "Protontricks",
            "--column", "Path"
        ]

    if len(steam_installations) == 1:
        return steam_installations[0]

    gui_provider = get_gui_provider()

    cmd_input = []

    for i, installation in enumerate(steam_installations):
        steam_path, steam_root = installation

        is_flatpak = (
            str(steam_path).endswith(
                "/com.valvesoftware.Steam/.local/share/Steam"
            )
        )
        is_snap = any(
            str(steam_path).endswith(snap_dir)
            for snap_dir in SNAP_STEAM_DIRS
        )

        if is_flatpak:
            install_type = "Flatpak"
        elif is_snap:
            install_type = "Snap"
        else:
            install_type = "Native"

        cmd_input.append(f"{i+1}: {install_type} - {steam_path}")

    cmd_input = "\n".join(cmd_input)

    if gui_provider == "yad":
        args = _get_yad_args()
    elif gui_provider == "zenity":
        args = _get_zenity_args()

    try:
        result = _run_gui(args, input_=cmd_input)
        choice = result.stdout
    except CalledProcessError as exc:
        if exc.returncode in (1, 252):
            # YAD returns 252 when dialog is closed by pressing Esc
            # No installation was selected
            choice = b""
        else:
            raise RuntimeError(
                f"{gui_provider} returned an error. Stderr: {exc.stderr}"
            )

    if choice in (b"", b" \n"):
        return None, None

    choice = choice.decode("utf-8").split(":")[0]
    choice = int(choice) - 1

    return steam_installations[choice]


def select_steam_app_with_gui(steam_apps, steam_path, title=None):
    """
    Prompt the user to select a Proton-enabled Steam app from
    a dropdown list.

    Return the selected SteamApp
    """
    def _get_yad_args():
        return [
            "yad", "--list", "--no-headers", "--center",
            "--window-icon", "wine",
            # Disabling markup means we won't have to escape special characters
            "--no-markup",
            "--search-column", "2",
            "--print-column", "2",
            "--width", "600", "--height", "400",
            "--text", title,
            "--title", "Protontricks",
            "--column", "Icon:IMG",
            "--column", "Steam app"
        ]

    def _get_zenity_args():
        return [
            "zenity", "--list", "--hide-header",
            "--width", "600",
            "--height", "400",
            "--text", title,
            "--title", "Protontricks",
            "--column", "Steam app"
        ]

    if not title:
        title = "Select Steam app"

    gui_provider = get_gui_provider()

    if gui_provider == "yad":
        args = _get_yad_args()

        # YAD implementation has icons for app selection
        appid2icon = _get_appid2icon(steam_apps)

        cmd_input = [
            [
                str(appid2icon[app.appid]),
                f"{app.name}: {app.appid}"
            ]
            for app in steam_apps if app.is_windows_app
        ]
        # Flatten the list
        cmd_input = list(itertools.chain.from_iterable(cmd_input))
    else:
        args = _get_zenity_args()
        cmd_input = [
            f'{app.name}: {app.appid}' for app in steam_apps
            if app.is_windows_app
        ]

    cmd_input = "\n".join(cmd_input)

    try:
        result = _run_gui(args, input_=cmd_input)
        choice = result.stdout
    except CalledProcessError as exc:
        # TODO: Remove this hack once the bug has been fixed upstream
        # Newer versions of zenity have a bug that causes long dropdown choice
        # lists to crash the command with a specific message.
        # Since stdout still prints the correct value, we can safely ignore
        # this error.
        #
        # The error is usually the message
        # 'free(): double free detected in tcache 2', but it can vary
        # depending on the environment. Instead, check if the returncode
        # is -6
        #
        # Related issues:
        # https://github.com/Matoking/protontricks/issues/20
        # https://gitlab.gnome.org/GNOME/zenity/issues/7
        if exc.returncode == -6:
            logger.info("Ignoring zenity crash bug")
            choice = exc.stdout
        elif exc.returncode in (1, 252):
            # YAD returns 252 when dialog is closed by pressing Esc
            # No game was selected
            choice = b""
        else:
            raise RuntimeError(
                f"{gui_provider} returned an error. Stderr: {exc.stderr}"
            )

    if choice in (b"", b" \n"):
        print("No game was selected. Quitting...")
        sys.exit(1)

    appid = str(choice).rsplit(':')[-1]
    appid = ''.join(x for x in appid if x.isdigit())
    appid = int(appid)

    steam_app = next(
        app for app in steam_apps
        if app.appid == appid)
    return steam_app


def prompt_filesystem_access(paths, show_dialog=False):
    """
    Check whether Protontricks has access to the provided file system paths
    and prompt the user to grant access if necessary.

    :param show_dialog: Show a dialog. If disabled, just print the message
                        instead.
    """
    def _map_path(path):
        """
        Map path to a path to be added into the `flatpak override` command.
        This means adding a tilde slash if the path is inside the home
        directory.
        """
        home_dir = str(Path.home())
        path = str(path)

        if path.startswith(home_dir):
            path = f"~/{path[len(home_dir)+1:]}"

        return path

    config = get_config()

    inaccessible_paths = get_inaccessible_paths(paths)
    inaccessible_paths = set(map(str, inaccessible_paths))

    logger.debug(
        "Following inaccessible paths were found: %s", inaccessible_paths
    )

    # Check what paths the user has ignored previously
    ignored_paths = set(
        json.loads(config.get("Dialog", "DismissedPaths", "[]"))
    )

    logger.debug("Following paths have been ignored: %s", ignored_paths)

    # Remaining paths that are inaccessible and that haven't been dismissed
    # by the user
    remaining_paths = inaccessible_paths - ignored_paths

    if not remaining_paths:
        return None

    cmd_filesystem = " ".join([
        "--filesystem={}".format(shlex.quote(_map_path(path)))
        for path in remaining_paths
    ])

    # TODO: Showing a text dialog and asking user to manually run the command
    # is very janky. Replace this with a proper permission prompt when
    # Flatpak supports it.
    message = (
        "Protontricks does not appear to have access to the following "
        "directories:\n"
        f" {' '.join(remaining_paths)}\n"
        "\n"
        "To fix this problem, grant access to the required directories by "
        "copying the following command and running it in a terminal:\n"
        "\n"
        f"flatpak override --user {cmd_filesystem} "
        "com.github.Matoking.protontricks\n"
        "\n"
        "You will need to restart Protontricks for the settings to take "
        "effect."
    )

    if show_dialog:
        ignore = show_text_dialog(
            title="Protontricks",
            text=message,
            window_icon="wine",
            cancel_label="Close",
            ok_label="Ignore, don't ask again",
            add_cancel_button=True
        )

        if ignore:
            # If user clicked "Don't ask again", store the paths to ensure the
            # user isn't prompted again for these directories
            ignored_paths |= inaccessible_paths

            config.set(
                "Dialog", "DismissedPaths", json.dumps(list(ignored_paths))
            )

    logger.warning(message)



import configparser
import logging
import os
from pathlib import Path

logger = logging.getLogger("protontricks")


class Config:
    def __init__(self):
        self._parser = configparser.ConfigParser()
        self._path = Path(
            os.environ.get("XDG_CONFIG_HOME", Path.home() / ".config")
        ) / "protontricks" / "config.ini"

        try:
            content = self._path.read_text(encoding="utf-8")
            self._parser.read_string(content)
        except FileNotFoundError:
            pass

    def get(self, section, option, default=None):
        """
        Get the configuration value in the given section and its field
        """
        self._parser.setdefault(section, {})
        return self._parser[section].get(option, default)

    def set(self, section, option, value):
        """
        Set the configuration value in the given section and its field, and
        save the configuration file
        """
        logger.debug(
            "Setting configuration field [%s][%s] = %s",
            section, option, value
        )
        self._parser.setdefault(section, {})
        self._parser[section][option] = value

        # Ensure parent directories exist
        self._path.parent.mkdir(parents=True, exist_ok=True)

        with self._path.open("wt", encoding="utf-8") as file_:
            self._parser.write(file_)


def get_config():
    """
    Retrieve the Protontricks configuration file
    """
    return Config()



from protontricks.config import get_config


def test_config(home_dir):
    """
    Test creating a configuration file, inserting a value into it and reading
    it back
    """
    config = get_config()

    config.set("General", "test_field", "test_value")

    # Ensure the configuration file now exists
    config_path = home_dir / ".config/protontricks/config.ini"
    assert config_path.exists()
    assert "test_value" in config_path.read_text()

    # Open the configuration file again, we should be able to read the value
    # back
    config = get_config()

    assert config.get("General", "test_field") == "test_value"


def test_config_default():
    """
    Test that a default value can be used if the field doesn't exist
    in the configuration file
    """
    config = get_config()

    assert config.get(
        "General", "fake_field", "default_value"
    ) == "default_value"




import locale
import logging
import os
import shlex
import shutil
import stat
import tempfile
from pathlib import Path
from subprocess import DEVNULL, PIPE, Popen, TimeoutExpired, check_output, run

import pkg_resources

__all__ = (
    "SUPPORTED_STEAM_RUNTIMES", "OS_RELEASE_PATHS", "lower_dict",
    "is_steam_deck", "get_legacy_runtime_library_paths",
    "get_host_library_paths", "RUNTIME_ROOT_GLOB_PATTERNS",
    "get_runtime_library_paths", "WINE_SCRIPT_TEMPLATE",
    "get_cache_dir", "create_wine_bin_dir", "run_command"
)

logger = logging.getLogger("protontricks")

SUPPORTED_STEAM_RUNTIMES = [
    # Old names
    "Steam Linux Runtime - Soldier",
    "Steam Linux Runtime - Sniper",

    # New names
    "Steam Linux Runtime 2.0 (soldier)",
    "Steam Linux Runtime 3.0 (sniper)"
]

OS_RELEASE_PATHS = [
    "/run/host/os-release",  # The host file if we're inside a Flatpak sandbox
    "/etc/os-release"
]


def lower_dict(d):
    """
    Return a copy of the dictionary with all keys recursively converted to
    lowercase.

    This is mainly used when dealing with Steam VDF files, as those tend to
    have either CamelCase or lowercase keys depending on the version.
    """
    def _lower_value(value):
        if not isinstance(value, dict):
            return value

        return {k.lower(): _lower_value(v) for k, v in value.items()}

    return {k.lower(): _lower_value(v) for k, v in d.items()}


def is_steam_deck():
    """
    Check if we're running on a Steam Deck
    """
    for path in OS_RELEASE_PATHS:
        try:
            lines = Path(path).read_text("utf-8").split("\n")
        except FileNotFoundError:
            continue

        if "ID=steamos" in lines and "VARIANT_ID=steamdeck" in lines:
            logger.info("The current device is a Steam Deck")
            return True

    return False


def get_legacy_runtime_library_paths(legacy_steam_runtime_path, proton_app):
    """
    Get LD_LIBRARY_PATH value to use when running a command using Steam Runtime
    """
    steam_runtime_paths = check_output([
        str(legacy_steam_runtime_path / "run.sh"),
        "--print-steam-runtime-library-paths"
    ])
    steam_runtime_paths = str(steam_runtime_paths, "utf-8")
    # Add Proton installation directory first into LD_LIBRARY_PATH
    # so that libwine.so.1 is picked up correctly (see issue #3)
    return "".join([
        str(proton_app.proton_dist_path / "lib"), os.pathsep,
        str(proton_app.proton_dist_path / "lib64"), os.pathsep,
        steam_runtime_paths
    ])


def get_host_library_paths():
    """
    Get host library paths to use when creating the LD_LIBRARY_PATH environment
    variable for use with newer Steam Runtime installations when *not*
    using bwrap
    """
    # The traditional Steam Runtime does the following when running the
    # `run.sh --print-steam-runtime-library-paths` command.
    # Since that command is unavailable with newer Steam Runtime releases,
    # do it ourselves here.
    result = run(
        ["/sbin/ldconfig", "-XNv"],
        check=True, stdout=PIPE, stderr=PIPE
    )
    lines = result.stdout.decode("utf-8").split("\n")
    paths = [
        line.split(":")[0] for line in lines
        if line.startswith("/") and ":" in line
    ]

    return ":".join(paths)


RUNTIME_ROOT_GLOB_PATTERNS = (
    "var/*/files/",
    "*/files/"
)


def get_runtime_library_paths(proton_app, use_bwrap=True):
    """
    Get LD_LIBRARY_PATH value to use when running a command using Steam Runtime
    """
    def find_runtime_app_root(runtime_app):
        """
        Find the runtime root (the directory containing the root fileystem
        used for the container) for separately installed Steam Runtime app
        """
        for pattern in RUNTIME_ROOT_GLOB_PATTERNS:
            try:
                return next(
                    runtime_app.install_path.glob(pattern)
                )
            except StopIteration:
                pass

        raise RuntimeError(
            f"Could not find Steam Runtime runtime root for {runtime_app.name}"
        )

    if use_bwrap:
        return "".join([
            str(proton_app.proton_dist_path / "lib"), os.pathsep,
            str(proton_app.proton_dist_path / "lib64"), os.pathsep
        ])

    runtime_root = find_runtime_app_root(proton_app.required_tool_app)
    return "".join([
        str(proton_app.proton_dist_path / "lib"), os.pathsep,
        str(proton_app.proton_dist_path / "lib64"), os.pathsep,
        get_host_library_paths(), os.pathsep,
        str(runtime_root / "lib" / "i386-linux-gnu"), os.pathsep,
        str(runtime_root / "lib" / "x86_64-linux-gnu")
    ])


WINE_SCRIPT_TEMPLATE = Path(
    pkg_resources.resource_filename(
        "protontricks", "data/scripts/wine_launch.sh"
    )
).read_text(encoding="utf-8")
WINESERVER_KEEPALIVE_SH_SCRIPT = Path(
    pkg_resources.resource_filename(
        "protontricks", "data/scripts/wineserver_keepalive.sh"
    )
).read_text(encoding="utf-8")
WINESERVER_KEEPALIVE_BATCH_SCRIPT = Path(
    pkg_resources.resource_filename(
        "protontricks", "data/scripts/wineserver_keepalive.bat"
    )
).read_text(encoding="utf-8")
BWRAP_LAUNCHER_SH_SCRIPT = Path(
    pkg_resources.resource_filename(
        "protontricks", "data/scripts/bwrap_launcher.sh"
    )
).read_text(encoding="utf-8")


def get_cache_dir():
    """
    Get Protontricks' cache directory, creating it first if it does not
    exist
    """
    xdg_cache_dir = os.environ.get(
        "XDG_CACHE_HOME", os.path.expanduser("~/.cache")
    )
    base_path = Path(xdg_cache_dir) / "protontricks"
    os.makedirs(str(base_path), exist_ok=True)

    return base_path


def create_wine_bin_dir(proton_app, use_bwrap=True):
    """
    Create a directory with "proxy" executables that load shared libraries
    using Steam Runtime and Proton's own libraries instead of the system
    libraries
    """
    binaries = list((proton_app.proton_dist_path / "bin").iterdir())

    # Create the base directory containing files for every Proton installation
    base_path = get_cache_dir() / "proton"
    os.makedirs(str(base_path), exist_ok=True)

    # Create a directory to hold the new executables for the specific
    # Proton installation
    bin_path = base_path / proton_app.name / "bin"
    bin_path.mkdir(parents=True, exist_ok=True)

    logger.info(
        "Created Steam Runtime Wine binary directory at %s", str(bin_path)
    )

    # Delete the directory and rewrite the scripts. Some binaries may no
    # longer exist in the Proton installation, so we'll also get rid of
    # scripts that point to non-existing files
    shutil.rmtree(str(bin_path))
    bin_path.mkdir(parents=True)

    for binary in binaries:
        proxy_script_path = bin_path / binary.name

        content = WINE_SCRIPT_TEMPLATE.replace(
            "@@name@@", shlex.quote(binary.name),
        )
        content = content.replace(
            "@@script_path@@", str(proxy_script_path)
        )

        proxy_script_path.write_text(content, encoding="utf-8")

        script_stat = proxy_script_path.stat()
        # Make the helper script executable
        proxy_script_path.chmod(script_stat.st_mode | stat.S_IEXEC)

    # Create the wineserver keepalive batch script
    (bin_path / "wineserver-keepalive.bat").write_text(
        WINESERVER_KEEPALIVE_BATCH_SCRIPT
    )
    keepalive_shell_script = bin_path / "wineserver-keepalive"
    keepalive_shell_script.write_text(
        WINESERVER_KEEPALIVE_SH_SCRIPT.replace(
            "@@keepalive_bat_path@@",
            str(bin_path / "wineserver-keepalive.bat")
        )
    )
    keepalive_shell_script.chmod(
        keepalive_shell_script.stat().st_mode | stat.S_IEXEC
    )
    launcher_script = bin_path / "bwrap-launcher"
    launcher_script.write_text(BWRAP_LAUNCHER_SH_SCRIPT)
    launcher_script.chmod(launcher_script.stat().st_mode | stat.S_IEXEC)

    return bin_path


def _get_fixed_locale_env():
    """Return a dictionary of fixed locale environment variables if Steam Deck
    is in use and some of the selected locales haven't actually been generated
    by the system.

    If the locale settings require no changes, an empty dict will be returned
    instead.
    """
    # We can assume the 'en_US.UTF-8' locale always exists on Steam Deck, but
    # we can't assume the same about other distros. Therefore, only attempt
    # fixing the locale when running on a Steam Deck.
    if not is_steam_deck():
        return {}

    supported_locales = run(
        ["locale", "-a"], check=True, stdout=PIPE, stderr=DEVNULL
    ).stdout.decode("utf-8").splitlines()
    supported_locales = [
        locale.normalize(locale_) for locale_
        in supported_locales
    ]

    locale_output = run(
        ["locale"], check=True, stdout=PIPE, stderr=DEVNULL
    ).stdout.decode("utf-8").splitlines()
    locale_output = [value.split("=") for value in locale_output]
    locale_settings = {
        value[0]: value[1].strip('"') for value in locale_output
    }

    fixed_env = {}

    # Check if any of the locales don't actually exist
    for category, locale_ in locale_settings.items():
        if locale_.strip() == "":
            continue

        # Normalize the locale name
        locale_ = locale.normalize(locale_)
        if locale_ not in supported_locales:
            # Locale does not exist
            fixed_env[category] = "en_US.UTF-8"

    if fixed_env:
        logger.warning(
            "Found locale categories configured with missing locales. "
            "The locale has been reset to 'en_US.UTF-8' for the "
            "following categories: %s",
            ", ".join(fixed_env.keys())
        )

    return fixed_env


def _start_process(args, wait=False, **kwargs):
    """Start a new process and return a Popen instance
    """
    process = Popen(args=args, **kwargs)

    if wait:
        process.wait()

    return process


def run_command(
        winetricks_path, proton_app, steam_app, command,
        use_steam_runtime=False,
        legacy_steam_runtime_path=None,
        use_bwrap=None,
        start_wineserver=None,
        env=None,
        **kwargs):
    """Run an arbitrary command with the correct environment variables
    for the given Proton app

    The environment variables are set for the duration of the call
    and restored afterwards

    If 'use_steam_runtime' is True, run the command using Steam Runtime
    using either 'legacy_steam_runtime_path' or the Proton app's specific
    Steam Runtime installation, depending on which one is required.

    If 'use_bwrap' is True, run newer Steam Runtime installations using
    bwrap based containerization. If None, determine whether bwrap is available
    and use it if so.

    If 'start_wineserver' is True, launch a background wineserver and keep it
    alive for the duration of the Protontricks call. If None, launch background
    wineserver if bwrap can be enabled.

    :returns: Return code of the executed command
    """
    # Check for incomplete Steam Runtime installation
    runtime_install_incomplete = \
        proton_app.required_tool_appid and not proton_app.required_tool_app

    if use_steam_runtime and runtime_install_incomplete:
        raise RuntimeError(
            f"{proton_app.name} is missing the required Steam Runtime. "
            "You may need to launch a Steam app using this Proton version "
            "to finish the installation."
        )

    # Make a copy of the environment variables to use for the subprocesses.
    # Include any additional environment variables if provided.
    if env is None:
        env = {}

    wine_environ = os.environ.copy()
    wine_environ.update(env)

    user_provided_wine = os.environ.get("WINE", False)
    user_provided_wineserver = os.environ.get("WINESERVER", False)

    wine_environ["WINETRICKS"] = str(winetricks_path)
    wine_environ["WINEPREFIX"] = str(steam_app.prefix_path)
    wine_environ["WINEDLLPATH"] = "".join([
        str(proton_app.proton_dist_path / "lib64" / "wine"),
        os.pathsep,
        str(proton_app.proton_dist_path / "lib" / "wine")
    ])

    wine_environ["PATH"] = "".join([
        str(proton_app.proton_dist_path / "bin"), os.pathsep,
        wine_environ["PATH"]
    ])

    # Expose the path to Proton installation. This is mainly used for
    # Wine helper scripts, but other scripts could use it as well.
    wine_environ["PROTON_PATH"] = str(proton_app.install_path)
    wine_environ["PROTON_DIST_PATH"] = str(proton_app.proton_dist_path)

    wine_environ["STEAM_APP_PATH"] = str(steam_app.install_path)
    wine_environ["STEAM_APPID"] = str(steam_app.appid)

    # Unset WINEARCH, which might be set for another Wine installation
    wine_environ.pop("WINEARCH", "")

    # Fix the locale for Steam Deck, if necessary
    wine_environ.update(_get_fixed_locale_env())

    wine_bin_dir = None
    wine_environ["PROTONTRICKS_STEAM_RUNTIME"] = "off"
    if use_steam_runtime:
        if use_bwrap is None:
            use_bwrap = bool(proton_app.required_tool_app)
            logger.info("Using 'bwrap = %s' as default value", use_bwrap)

        if start_wineserver is None:
            start_wineserver = use_bwrap
            logger.info(
                "Using 'background-wineserver = %s' as default value",
                start_wineserver
            )

        if proton_app.required_tool_app:
            wine_environ["STEAM_RUNTIME_PATH"] = \
                str(proton_app.required_tool_app.install_path)
            wine_environ["PROTON_LD_LIBRARY_PATH"] = \
                get_runtime_library_paths(proton_app, use_bwrap=use_bwrap)

            runtime_name = proton_app.required_tool_app.name
            logger.info(
                "Using separately installed Steam Runtime: %s",
                runtime_name
            )

            if use_bwrap:
                wine_environ["PROTONTRICKS_STEAM_RUNTIME"] = "bwrap"
                logger.info(
                    "Running Steam Runtime using bwrap containerization.\n"
                    "If any problems arise, please try running the command "
                    "again using the `--no-bwrap` flag and make an issue "
                    "report if the problem only occurs when bwrap is in use."
                )
            else:
                wine_environ["PROTONTRICKS_STEAM_RUNTIME"] = "legacy"

            if runtime_name not in SUPPORTED_STEAM_RUNTIMES:
                logger.warning(
                    "Current Steam Runtime not recognized by Protontricks."
                )
        else:
            # Legacy Steam Runtime requires a different LD_LIBRARY_PATH
            # that is produced by a script.
            wine_environ["PROTONTRICKS_STEAM_RUNTIME"] = "legacy"
            wine_environ["PROTON_LD_LIBRARY_PATH"] = \
                get_legacy_runtime_library_paths(
                    legacy_steam_runtime_path, proton_app
                )

            # bwrap is not available, so ensure it is not launched even if the
            # user configured it so
            use_bwrap = False

    # Configure the environment to use launch scripts that take care of
    # configuring the environment and Wine before launching the underlying
    # Wine binaries.
    wine_bin_dir = create_wine_bin_dir(proton_app)
    wine_environ["LEGACY_STEAM_RUNTIME_PATH"] = str(legacy_steam_runtime_path)
    wine_environ["PATH"] = os.pathsep.join(
        [str(wine_bin_dir), wine_environ["PATH"]]
    )

    if not user_provided_wine:
        logger.info(
            "WINE environment variable is not available. "
            "Setting WINE environment variable to Proton bundled version."
        )
        wine_environ["WINE"] = str(wine_bin_dir / "wine")

    wine_environ["WINELOADER"] = wine_environ["WINE"]

    if not user_provided_wineserver:
        logger.info(
            "WINESERVER environment variable is not available. "
            "Setting WINESERVER environment variable to Proton bundled version"
        )
        wine_environ["WINESERVER"] = str(wine_bin_dir / "wineserver")

    temp_dir = Path(tempfile.mkdtemp(prefix="protontricks-"))
    wine_environ["PROTONTRICKS_TEMP_PATH"] = str(temp_dir)
    wine_environ["PROTONTRICKS_SESSION_ID"] = temp_dir.name.split("-")[1]

    if start_wineserver:
        wine_environ["PROTONTRICKS_BACKGROUND_WINESERVER"] = "1"

    launcher_process = None
    keepalive_process = None
    try:
        if use_bwrap:
            logger.info(
                "Starting bwrap launcher process: %s",
                str(wine_bin_dir / "bwrap-launcher")
            )

            # TODO: Waiting for launcher to start can be simplified once
            # ValveSoftware/steam-runtime#593 has been fixed and stdout can
            # be used instead.
            launcher_read_fd, launcher_write_fd = os.pipe2(os.O_CLOEXEC)
            launcher_process = _start_process(
                [str(wine_bin_dir / "bwrap-launcher"), str(launcher_write_fd)],
                wait=False,
                pass_fds=[launcher_write_fd],
                env=wine_environ
            )

            # The Steam Runtime launcher service will write to the given
            # file descriptor and then close it to indicate the launcher is
            # ready or about to exit (i.e. due to wrong CLI parameters).
            os.close(launcher_write_fd)
            with open(launcher_read_fd, "rb") as reader:
                reader.read()

            # Check if the launcher actually started up and is still running.
            try:
                launcher_process.wait(timeout=0.1)
                # Launcher process crashed, bail out
                raise RuntimeError(
                    f"bwrap launcher crashed, "
                    f"returncode: {launcher_process.returncode}"
                )
            except TimeoutExpired:
                # Launcher is running as expected
                pass

            logger.info("bwrap launcher started")

        if start_wineserver:
            logger.info(
                "Starting wineserver keepalive process: %s",
                str(wine_bin_dir / "wineserver-keepalive")
            )
            keepalive_process = _start_process(
                str(wine_bin_dir / "wineserver-keepalive"),
                wait=False,
                env=wine_environ,
                stdout=DEVNULL,
            )

        logger.info("Attempting to run command %s", command)

        kwargs = kwargs.copy()
        kwargs["env"] = wine_environ

        process = _start_process(
            command, wait=True, **kwargs
        )
        return process.returncode
    finally:
        shutil.rmtree(str(temp_dir), ignore_errors=True)

        if keepalive_process:
            logger.info(
                "Terminating wineserver keepalive process %d",
                keepalive_process.pid
            )
            keepalive_process.terminate()

        if launcher_process:
            logger.info(
                "Terminating launcher process %d",
                launcher_process.pid
            )
            launcher_process.terminate()
            launcher_process.wait()
            logger.info("Launcher process terminated")



import pytest

from pathlib import Path

from protontricks.flatpak import (get_inaccessible_paths,
                                  get_running_flatpak_version)


class TestGetRunningFlatpakVersion:
    def test_flatpak_not_active(self):
        """
        Test Flatpak version detection when Flatpak is not active
        """
        assert get_running_flatpak_version() is None

    def test_flatpak_active(self, monkeypatch, tmp_path):
        """
        Test Flatpak version detection when Flatpak is active
        """
        flatpak_info_path = tmp_path / "flatpak-info"

        flatpak_info_path.write_text(
            "[Application]\n"
            "name=fake.flatpak.Protontricks\n"
            "\n"
            "[Instance]\n"
            "flatpak-version=1.12.1"
        )
        monkeypatch.setattr(
            "protontricks.flatpak.FLATPAK_INFO_PATH", str(flatpak_info_path)
        )

        assert get_running_flatpak_version() == (1, 12, 1)


class TestGetInaccessiblePaths:
    def test_flatpak_disabled(self):
        """
        Test that an empty list is returned if Flatpak is not active
        """
        assert get_inaccessible_paths(["/fake", "/fake_2"]) == []

    def test_flatpak_active(self, monkeypatch, home_dir, tmp_path):
        """
        Test that inaccessible paths are correctly detected when
        Flatpak is active
        """
        flatpak_info_path = tmp_path / "flatpak-info"

        flatpak_info_path.write_text(
            "[Application]\n"
            "name=fake.flatpak.Protontricks\n"
            "\n"
            "[Instance]\n"
            "flatpak-version=1.12.1\n"
            "\n"
            "[Context]\n"
            "filesystems=/mnt/SSD_A;/mnt/SSD_B;xdg-data/Steam;"
        )
        monkeypatch.setattr(
            "protontricks.flatpak.FLATPAK_INFO_PATH", str(flatpak_info_path)
        )

        inaccessible_paths = get_inaccessible_paths([
            "/mnt/SSD_A", "/mnt/SSD_C",
            str(home_dir / ".local/share/SteamOld"),
            str(home_dir / ".local/share/Steam")
        ])
        assert len(inaccessible_paths) == 2
        assert str(inaccessible_paths[0]) == "/mnt/SSD_C"
        assert str(inaccessible_paths[1]) == \
            str(Path("~/.local/share/SteamOld").expanduser())

    def test_flatpak_home(self, monkeypatch, tmp_path, home_dir):
        """
        Test that 'home' filesystem permission grants permission to the
        home directory
        """
        flatpak_info_path = tmp_path / "flatpak-info"

        flatpak_info_path.write_text(
            "[Application]\n"
            "name=fake.flatpak.Protontricks\n"
            "\n"
            "[Instance]\n"
            "flatpak-version=1.12.1\n"
            "\n"
            "[Context]\n"
            "filesystems=home;"
        )
        monkeypatch.setattr(
            "protontricks.flatpak.FLATPAK_INFO_PATH", str(flatpak_info_path)
        )

        inaccessible_paths = get_inaccessible_paths([
            "/mnt/SSD_A", "/var/fake_path",
            str(home_dir / "fake_path"),
            str(home_dir / ".local/share/FakePath")
        ])

        assert len(inaccessible_paths) == 2
        assert str(inaccessible_paths[0]) == "/mnt/SSD_A"
        assert str(inaccessible_paths[1]) == "/var/fake_path"

    def test_flatpak_home_tilde(self, monkeypatch, tmp_path, home_dir):
        """
        Test that tilde slash is expanded if included in the list of
        file systems
        """
        flatpak_info_path = tmp_path / "flatpak-info"

        flatpak_info_path.write_text(
            "[Application]\n"
            "name=fake.flatpak.Protontricks\n"
            "\n"
            "[Instance]\n"
            "flatpak-version=1.12.1\n"
            "\n"
            "[Context]\n"
            "filesystems=~/fake_path"
        )
        monkeypatch.setattr(
            "protontricks.flatpak.FLATPAK_INFO_PATH", str(flatpak_info_path)
        )

        inaccessible_paths = get_inaccessible_paths([
            str(home_dir / "fake_path"),
            str(home_dir / "fake_path_2")
        ])

        assert len(inaccessible_paths) == 1
        assert str(inaccessible_paths[0]) == str(home_dir / "fake_path_2")

    def test_flatpak_host(self, monkeypatch, tmp_path, home_dir):
        """
        Test that 'host' filesystem permission grants permission to the
        whole file system
        """
        flatpak_info_path = tmp_path / "flatpak-info"

        flatpak_info_path.write_text(
            "[Application]\n"
            "name=fake.flatpak.Protontricks\n"
            "\n"
            "[Instance]\n"
            "flatpak-version=1.12.1\n"
            "\n"
            "[Context]\n"
            "filesystems=host;"
        )
        monkeypatch.setattr(
            "protontricks.flatpak.FLATPAK_INFO_PATH", str(flatpak_info_path)
        )

        inaccessible_paths = get_inaccessible_paths([
            "/mnt/SSD_A", "/var/fake_path",
            str(home_dir / "fake_path"),
        ])

        assert len(inaccessible_paths) == 0

    @pytest.mark.usefixtures("xdg_user_dir_bin")
    def test_flatpak_xdg_user_dir(self, monkeypatch, tmp_path, home_dir):
        """
        Test that XDG filesystem permissions such as 'xdg-pictures' and
        'xdg-download' are detected correctly
        """
        flatpak_info_path = tmp_path / "flatpak-info"

        flatpak_info_path.write_text(
            "[Application]\n"
            "name=fake.flatpak.Protontricks\n"
            "\n"
            "[Instance]\n"
            "flatpak-version=1.12.1\n"
            "\n"
            "[Context]\n"
            "filesystems=xdg-pictures;"
        )
        monkeypatch.setattr(
            "protontricks.flatpak.FLATPAK_INFO_PATH", str(flatpak_info_path)
        )

        inaccessible_paths = get_inaccessible_paths([
            str(home_dir / "Pictures"),
            str(home_dir / "Download")
        ])

        assert len(inaccessible_paths) == 1
        assert str(inaccessible_paths[0]) == str(home_dir / "Download")

    def test_flatpak_unknown_permission(self, monkeypatch, tmp_path, caplog):
        """
        Test that unknown filesystem permissions are ignored
        """
        flatpak_info_path = tmp_path / "flatpak-info"

        flatpak_info_path.write_text(
            "[Application]\n"
            "name=fake.flatpak.Protontricks\n"
            "\n"
            "[Instance]\n"
            "flatpak-version=1.12.1\n"
            "\n"
            "[Context]\n"
            "filesystems=home;unknown-fs;"
        )
        monkeypatch.setattr(
            "protontricks.flatpak.FLATPAK_INFO_PATH", str(flatpak_info_path)
        )

        inaccessible_paths = get_inaccessible_paths([
            "/mnt/SSD",
        ])

        assert len(inaccessible_paths) == 1

        # Unknown filesystem permission is logged
        records = caplog.records

        assert len(records) == 1
        assert records[0].levelname == "WARNING"
        assert "Unknown Flatpak file system permission 'unknown-fs'" \
            in records[0].message



import logging
import os
import string
import struct
import zlib
from collections import OrderedDict
from pathlib import Path

import vdf

from ._vdf import binary_loads as vendored_binary_loads
from .util import is_steam_deck, lower_dict

__all__ = (
    "COMMON_STEAM_DIRS", "SteamApp", "find_steam_installations",
    "find_steam_path", "find_legacy_steam_runtime_path",
    "iter_appinfo_sections", "get_appinfo_sections", "get_tool_appid",
    "find_steam_compat_tool_app", "find_appid_proton_prefix",
    "find_proton_app", "get_steam_lib_paths", "get_compat_tool_dirs",
    "get_custom_compat_tool_installations_in_dir",
    "get_custom_compat_tool_installations", "find_current_steamid3",
    "get_appid_from_shortcut", "get_custom_windows_shortcuts",
    "get_steam_apps"
)

COMMON_STEAM_DIRS = [
    ".steam/steam",
    ".local/share/Steam",
]

SNAP_STEAM_DIRS = [
    "snap/steam/common/.local/share/Steam",

    # Location if 'experimental.hidden-snap-folder' Snap feature flag is set
    ".snap/data/steam/common/.local/share/Steam",
]

OS_RELEASE_PATHS = [
    "/run/host/os-release",  # The host file if we're inside a Flatpak sandbox
    "/etc/os-release"
]

# 'cache/appinfo.vdf' contains a VDF section with metadata about
# "Proton name -> app ID" mappings with this app ID
STEAM_PLAY_MANIFESTS_APPID = 891390

logger = logging.getLogger("protontricks")


class SteamApp(object):
    """
    SteamApp represents an installed Steam app or whatever is close enough to
    one (eg. a custom Proton installation or a Windows shortcut with its own
    Proton prefix)
    """
    __slots__ = (
        "appid", "name", "prefix_path", "install_path", "icon_path",
        "required_tool_appid", "required_tool_app"
    )

    def __init__(
            self, name, install_path, icon_path=None, prefix_path=None,
            appid=None, required_tool_appid=None):
        """
        :appid: App's appid
        :name: The app's human-readable name
        :prefix_path: Absolute path to where the app's Wine prefix *might*
                      exist.
        :app_path: Absolute path to app's installation directory
        :icon_path: Absolute path to app's icon, whether Steam's own or custom
                    one configured by user
        :required_tool_appid: App ID required to run this application.
                              Usually corresponds to a Steam Runtime for
                              Proton installations.
        """
        self.appid = int(appid) if appid else None
        self.required_tool_appid = \
            int(required_tool_appid) if required_tool_appid else None

        self.name = name

        if prefix_path:
            self.prefix_path = Path(prefix_path)
        else:
            self.prefix_path = None

        if icon_path:
            self.icon_path = Path(icon_path)
        else:
            self.icon_path = None

        self.install_path = Path(install_path)

        # Reference to another SteamApp will be added later if necessary,
        # once we have the full list of Steam apps
        self.required_tool_app = None

    @property
    def prefix_path_exists(self):
        """
        Returns True if the app has a Wine prefix directory that has been
        launched at least once
        """
        if not self.prefix_path:
            return False

        # 'pfx' directory is incomplete until the game has been launched
        # once, so check for 'pfx.lock' as well
        return (
            self.prefix_path.is_dir()
            and (self.prefix_path.parent / "pfx.lock").is_file()
        )

    def name_contains(self, s):
        """
        Returns True if the name contains the given substring.
        Both strings are normalized for easier searching before comparison.
        """
        def normalize_str(s):
            """
            Normalize the string to make it easier for human to
            perform a search by removing all symbols
            except ASCII digits and letters and turning it into lowercase
            """
            printable = set(string.printable) - set(string.punctuation)
            s = "".join([c for c in s if c in printable])
            s = s.lower()
            s = s.replace(" ", "")
            return s

        return normalize_str(s) in normalize_str(self.name)

    @property
    def is_proton(self):
        """
        Return True if this app is a Proton installation
        """
        # If the installation directory contains a file named "proton",
        # it's a Proton installation
        return (self.install_path / "proton").is_file()

    @property
    def is_tool(self):
        """
        Return True if this app is a tool rather an app.

        This is true for Proton and Steam Runtime installations.
        """
        return (self.install_path / "toolmanifest.vdf").is_file()

    @property
    def is_windows_app(self):
        """
        Return True if this app is a Windows app that's launched using Proton
        """
        return not self.is_proton and self.prefix_path_exists and self.appid

    @property
    def is_proton_ready(self):
        """
        Return True if the Proton installation is ready for use.

        Proton installation might be incomplete if it hasn't been launched
        yet, in which case the Proton binaries don't exist yet.
        """
        return bool(self.proton_dist_path)

    @property
    def proton_dist_path(self):
        """
        Return path to the directory containing Proton binaries and libraries.
        None is returned if this app isn't a Proton installation or either
        directory doesn't exist.

        The directory is named either 'dist' or 'files'.

        'dist' is used by older Proton releases, and it is extracted from a
        separate 'proton_dist.tar' archive during first launch.
        'files' is used by newer Proton releases, and it already exists
        after the Steam app has been installed, requiring no first launch.
        """
        if not self.is_proton:
            return None

        try:
            # Prioritize 'files' directory if it exists.
            # If both directories exist, 'dist' is likely a leftover that
            # wasn't removed by Steam.
            return next(
                (self.install_path / name) for name in ("files", "dist")
                if (self.install_path / name).is_dir()
            )
        except StopIteration:
            return None

    @classmethod
    def from_appmanifest(cls, path, steam_lib_paths, steam_path=None):
        """
        Parse appmanifest_X.acf file containing Steam app installation metadata
        and return a SteamApp object

        If 'steam_path' is provided, icon path is also populated
        """
        logger.debug(
            "Creating SteamApp from manifest file in %s", path
        )

        try:
            content = path.read_text(encoding="utf-8")
        except UnicodeDecodeError:
            # This might occur if the appmanifest becomes corrupted
            # eg. due to running a Linux filesystem under Windows
            # In that case just skip it
            logger.warning(
                "Skipping malformed appmanifest %s", path
            )
            return None
        except PermissionError:
            # Skip the appmanifest if we can't read it.
            # Steam also seems to ignore unreadable app manifests, so do the
            # same here.
            logger.warning(
                "Skipping appmanifest %s due to insufficient permissions",
                path
            )
            return None

        try:
            vdf_data = lower_dict(vdf.loads(content))
        except SyntaxError:
            logger.warning("Skipping malformed appmanifest %s", path)
            return None

        try:
            app_state = vdf_data["appstate"]
        except KeyError:
            # Some appmanifest files may be empty. Ignore those.
            logger.info("Skipping empty appmanifest %s", path)
            return None

        # The app ID field can be named 'appID' or 'appid'.
        # 'appid' is more common, but certain appmanifest
        # files (created by old Steam clients?) also use 'appID'.
        #
        # Use case-insensitive field names to deal with these.
        app_state = lower_dict(app_state)
        appid = int(app_state["appid"])

        try:
            name = app_state["name"]
        except KeyError:
            # Older app installations also use `userconfig/name`
            name = app_state["userconfig"]["name"]

        # Proton prefix may exist on a different library
        prefix_path = find_appid_proton_prefix(
            appid=appid, steam_lib_paths=steam_lib_paths
        )

        install_path = Path(path).parent / "common" / app_state["installdir"]

        icon_path = None

        # If Steam path was provided, also populate the icon path
        if steam_path:
            icon_path = \
                steam_path / "appcache" / "librarycache" / f"{appid}_icon.jpg"

        # Check if the app requires another app. This is the case with
        # newer versions of Proton, which use Steam Runtimes installed as
        # normal Steam apps
        try:
            required_tool_appid = _get_required_tool_appid(install_path)
        except (ValueError, SyntaxError):
            logger.warning(
                "Tool manifest for %s is empty or corrupted. You may need to "
                "reinstall the application.",
                name
            )
            return None

        return cls(
            appid=appid, name=name, prefix_path=prefix_path,
            install_path=install_path, icon_path=icon_path,
            required_tool_appid=required_tool_appid
        )


def _get_required_tool_appid(path):
    """
    Get the required tool app ID for the Proton installation at the given path

    :raises ValueError: Tool manifest is empty
    :raises SyntaxError: Tool manifest is corrupted
    """
    tool_manifest_path = path / "toolmanifest.vdf"
    try:
        tool_manifest_content = tool_manifest_path.read_text()

        if tool_manifest_content == "":
            raise ValueError("Tool manifest is empty")

        tool_manifest = lower_dict(vdf.loads(tool_manifest_content))

        return tool_manifest["manifest"].get("require_tool_appid", None)
    except FileNotFoundError:
        return None


def _get_steamapps_subdirs(path):
    """
    Get all directories under the given path that match 'steamapps'
    in a case-insensitive manner
    """
    try:
        dirs = list(
            entry for entry in path.iterdir()
            if entry.name.lower() == "steamapps" and entry.is_dir()

        )
    except FileNotFoundError:
        # Directory does not exist
        return []

    # Sort entries so that 'steamapps' is listed first, as it's the default
    # directory name that Steam uses and should thus be prioritized
    dirs = list(reversed(sorted(dirs, key=lambda dir_: dir_.name)))

    return dirs


def find_steam_installations():
    """
    Find all Steam installations and return them as a list of (steam_path, steam_root)
    tuples
    """
    def has_steamapps_dir(path):
        """
        Return True if the path either has a 'steamapps' subdirectory,
        False otherwise
        """
        # Steam doesn't care about case-insensitivity and considers even
        # names like 'SteaMAPps' valid
        try:
            return any(
                path.name.lower() == "steamapps" and path.is_dir() for path
                in path.iterdir()
            )
        except FileNotFoundError:
            # Directory does not exist
            return False

    def has_runtime_dir(path):
        return (path / "ubuntu12_32").is_dir()

    # as far as @admalledd can tell,
    # this should always be correct for the tools root:
    steam_root = (Path.home() / ".steam" / "root").resolve()

    if not (steam_root / "ubuntu12_32").is_dir():
        # Check that runtime dir exists, if not make root=path and hope
        steam_root = None

    if os.environ.get("STEAM_DIR"):
        steam_path = Path(os.environ.get("STEAM_DIR"))
        if has_steamapps_dir(steam_path) and has_runtime_dir(steam_path):
            logger.info(
                "Found a valid Steam installation at %s.", steam_path
            )

            return [
                (Path(steam_path), Path(steam_path))
            ]

        logger.error(
            "$STEAM_DIR was provided but didn't point to a valid Steam "
            "installation."
        )

        return []

    # Track the found Steam directory candidates using an ordered dict,
    # ensuring that any duplicates are eliminated and that we pick the first
    # candidate we find.
    # We essentially use this as an ordered set.
    candidates = OrderedDict()

    for steam_path in COMMON_STEAM_DIRS:
        # The common Steam directories are found inside the home directory
        steam_path = (Path.home() / steam_path).resolve()
        if has_steamapps_dir(steam_path):
            if not steam_root:
                steam_root_ = steam_path
            else:
                steam_root_ = steam_root

            candidates[(str(steam_path), str(steam_root_))] = True

    # Check for Flatpak and Snap Steam separately and ensure we don't mix steam_root
    # and steam_path from Flatpak/Snap and native installations of Steam.
    steam_path = \
        Path.home() / ".var/app/com.valvesoftware.Steam/data/Steam"
    steam_path = steam_path.resolve()
    if has_steamapps_dir(steam_path):
        candidates[(str(steam_path), str(steam_path))] = True

    for steam_dir in SNAP_STEAM_DIRS:
        steam_path = (Path.home() / steam_dir).resolve()
        if has_steamapps_dir(steam_path):
            candidates[(str(steam_path), str(steam_path))] = True

    for steam_path, _ in candidates.keys():
        logger.info("Found Steam directory at %s", steam_path)

    return [
        [Path(steam_path), Path(steam_root)]
        for steam_path, steam_root in candidates.keys()
    ]


def find_steam_path():
    """
    Try to discover default Steam dir using common locations and return the
    first one that matches

    Return (steam_path, steam_root), where steam_path points to
    "~/.steam/steam" (contains "appcache", "config" and "steamapps")
    and "~/.steam/root" (contains "ubuntu12_32" and "compatibilitytools.d")
    """
    candidates = find_steam_installations()

    if len(candidates) > 1:
        logger.warning(
            "Found multiple Steam directories. If you want to select a "
            "specific installation, use STEAM_DIR environment variable to set "
            "the correct directory:"
        )
        logger.warning("$ STEAM_DIR=<path> protontricks <appid> <command>")
        logger.warning("The following Steam directories were found:")
        for steam_path, _ in candidates:
            logger.warning(" %s", str(steam_path))

    try:
        steam_path, steam_root = candidates[0]
        logger.info(
            "Using Steam directory at %s. You can also define Steam directory "
            "manually using $STEAM_DIR", str(steam_path)
        )
        return Path(steam_path), Path(steam_root)
    except IndexError:
        return None, None


def find_legacy_steam_runtime_path(steam_root):
    """
    Find the legacy Steam Runtime either using the STEAM_RUNTIME env or
    steam_root
    """
    env_steam_runtime = os.environ.get("STEAM_RUNTIME", "")

    if env_steam_runtime == "0":
        # User has disabled Steam Runtime
        logger.info("STEAM_RUNTIME is 0. Disabling Steam Runtime.")
        return None
    elif env_steam_runtime and Path(env_steam_runtime).is_dir():
        # User has a custom Steam Runtime
        logger.info(
            "Using custom Steam Runtime at %s", env_steam_runtime)
        return Path(env_steam_runtime)
    elif env_steam_runtime in ["1", ""]:
        # User has enabled Steam Runtime or doesn't have STEAM_RUNTIME set;
        # default to enabled Steam Runtime in either case
        steam_runtime_path = steam_root / "ubuntu12_32" / "steam-runtime"

        logger.info(
            "Using default Steam Runtime at %s", str(steam_runtime_path))
        return steam_runtime_path

    logger.error(
        "Path in STEAM_RUNTIME doesn't point to a valid Steam Runtime!")

    return None


APPINFO_STRUCT_HEADER = "<4sL"
APPINFO_V28_STRUCT_SECTION = "<LLLLQ20sL20s"
APPINFO_V29_STRUCT_SECTION = "<LLLLQ20sL20s"


def iter_appinfo_sections(path):
    """
    Parse an appinfo.vdf file and iterate through all the binary VDF objects
    inside it
    """
    def _iter_v28_appinfo(data, start):
        """
        Parse and iterate appinfo.vdf version 28.
        Identical to version 27 except for new VDF SHA1 field.
        """
        i = start
        section_size = struct.calcsize(APPINFO_V28_STRUCT_SECTION)
        while True:
            # We don't need any of the fields besides 'entry_size',
            # which is used to determine the length of the variable-length VDF
            # field.
            # Still, here they are for posterity's sake.
            (appid, entry_size, infostate, last_updated, access_token,
             sha_hash, change_number, vdf_sha_hash) = struct.unpack(
                APPINFO_V28_STRUCT_SECTION, data[i:i+section_size])
            vdf_section_size = entry_size - (section_size - 8)

            i += section_size

            vdf_d = vdf.binary_loads(data[i:i+vdf_section_size])
            vdf_d = lower_dict(vdf_d)
            yield vdf_d

            i += vdf_section_size

            if i == len(data) - 4:
                return

    def _iter_v29_appinfo(data, start):
        """
        Parse and iterate appinfo.vdf version 29.
        """
        i = start

        # The header contains the offset to the key table
        key_table_offset = struct.unpack("<q", data[i:i+8])[0]
        key_table = []

        key_count = struct.unpack(
            "<i", data[key_table_offset:key_table_offset+4]
        )[0]

        table_i = key_table_offset + 4
        for _ in range(0, key_count):
            key = bytearray()
            while True:
                key.append(data[table_i])
                table_i += 1

                if key[-1] == 0:
                    key_table.append(
                        key[0:-1].decode("utf-8", errors="replace")
                    )
                    break

        i += 8

        section_size = struct.calcsize(APPINFO_V29_STRUCT_SECTION)
        while True:
            # We don't need any of the fields besides 'entry_size',
            # which is used to determine the length of the variable-length VDF
            # field.
            # Still, here they are for posterity's sake.
            (appid, entry_size, infostate, last_updated, access_token,
             sha_hash, change_number, vdf_sha_hash) = struct.unpack(
                APPINFO_V29_STRUCT_SECTION, data[i:i+section_size])
            vdf_section_size = entry_size - (section_size - 8)

            i += section_size

            try:
                vdf_d = vdf.binary_loads(
                    data[i:i+vdf_section_size], key_table=key_table
                )
            except TypeError:
                # System 'vdf' is too old and does not support 'key_table',
                # use the bundled one instead. This is cursed, but it's
                # so far the only reasonable option without a proper maintained
                # release on PyPI.
                vdf_d = vendored_binary_loads(
                    data[i:i+vdf_section_size], key_table=key_table
                )

            vdf_d = lower_dict(vdf_d)
            yield vdf_d

            i += vdf_section_size

            if i == key_table_offset - 4:
                return

    logger.debug("Loading appinfo.vdf in %s", path)

    # appinfo.vdf is not actually a (binary) VDF file, but a binary file
    # containing multiple binary VDF sections.
    # File structure based on comment from vdf developer:
    # https://github.com/ValvePython/vdf/issues/13#issuecomment-321700244
    data = path.read_bytes()
    i = 0

    # Parse the header
    header_size = struct.calcsize(APPINFO_STRUCT_HEADER)
    magic, universe = struct.unpack(
        APPINFO_STRUCT_HEADER, data[0:header_size]
    )

    i += header_size

    logger.debug("appinfo.vdf has magic number %s", magic)

    if magic == b'(DV\x07':
        yield from _iter_v28_appinfo(data, i)
    elif magic == b')DV\x07':
        yield from _iter_v29_appinfo(data, i)
    else:
        raise SyntaxError(
            "Invalid file magic number. The appinfo.vdf version might not be "
            "supported by the current version of Protontricks - please "
            "check for updates."
        )


def get_appinfo_sections(path):
    """
    Parse an appinfo.vdf file and return all the deserialized binary VDF
    objects inside it
    """
    return list(iter_appinfo_sections(path))


def get_tool_appid(compat_tool_name, steam_play_manifest):
    """
    Get the App ID for compatibility tool by the compat tool name
    used in STEAM_DIR/config/config.vdf
    """
    compat_tools = steam_play_manifest["appinfo"]["extended"]["compat_tools"]

    for default_name, entry in compat_tools.items():
        # A single compatibility tool may have multiple valid names
        # eg. "proton_316" and "proton_316_beta"
        aliases = [default_name]

        # Each compat tool entry can also contain an 'aliases' field
        # with a different compat tool name
        if "aliases" in entry:
            aliases += entry["aliases"].split(",")

        logger.debug("%s has compat tool aliases %s", default_name, aliases)

        if compat_tool_name in aliases:
            return entry["appid"]

    return None


def find_steam_compat_tool_app(steam_path, steam_apps, appid=None):
    """
    Get the current compatibility tool used by Steam and
    return a SteamApp object

    If 'appid' is provided, try to find the app-specific compatibility tool
    if one is configured

    The compatibility tool *may* not be a Proton installation. This can be
    checked using `SteamApp.is_proton`.
    """
    def _get_tool_app(compat_tool_name, steam_apps, steam_play_manifest):
        tool_appid = get_tool_appid(compat_tool_name, steam_play_manifest)

        if not tool_appid:
            return None

        try:
            app = next(
                app for app in steam_apps if app.appid == tool_appid
            )
            return app
        except StopIteration:
            return None

    logger.debug("Finding Steam compat tool name for appid %s", appid)

    # Determine which Proton version to use (either globally or for
    # a specific app if `appid` was provided) by checking
    # `<steam_dir>/config/config.vdf` and `<steam_dir>/appcache/appinfo.vdf`.
    # Multiple configuration values might be found. In such case, select the
    # one with the highest priority.
    config_vdf_path = steam_path / "config" / "config.vdf"
    # config.vdf might contain invalid UTF-8 characters in the
    # `SDL_GamepadBind` field. We don't use that in any way, so we can deal
    # with the invalid characters by just ignoring them.
    content = config_vdf_path.read_text(errors="replace")

    vdf_data = lower_dict(vdf.loads(content))

    appinfo_path = steam_path / "appcache" / "appinfo.vdf"
    appinfo_sections = [
        section for section in iter_appinfo_sections(appinfo_path)
        if section["appinfo"]["appid"] in (STEAM_PLAY_MANIFESTS_APPID, appid)
    ]
    steam_play_manifest = next(
        section for section in appinfo_sections
        if section["appinfo"]["appid"] == STEAM_PLAY_MANIFESTS_APPID
    )

    try:
        app_section = next(
            section for section in appinfo_sections
            if section["appinfo"]["appid"] == appid
        )
    except StopIteration:
        # App ID was most likely not provided
        app_section = None

    try:
        manifest_app_compat_section = next(
            entry for mapping_appid, entry in
            steam_play_manifest["appinfo"]["extended"]["app_mappings"].items()
            if int(mapping_appid) == appid
        )
    except StopIteration:
        # App doesn't have a default compatibility tool mapping
        manifest_app_compat_section = None

    # ToolMapping seems to be used in older Steam beta releases
    try:
        tool_mapping = (
            vdf_data["installconfigstore"]["software"]["valve"]["steam"]
                    ["toolmapping"]
        )
        logger.debug("Found ToolMapping entry")
    except KeyError:
        tool_mapping = {}

    # CompatToolMapping seems to be the name used in newer Steam releases
    # We'll prioritize this if it exists
    try:
        compat_tool_mapping = (
            vdf_data["installconfigstore"]["software"]["valve"]["steam"]
                    ["compattoolmapping"]
        )
        logger.debug("Found CompatToolMapping entry")
    except KeyError:
        compat_tool_mapping = {}

    # The name of potential names in order of priority
    potential_names = []

    # Game specific user settings have the 1st priority, if they exist
    if compat_tool_mapping.get(str(appid), {}).get("name"):
        tool_name = compat_tool_mapping[str(appid)]["name"]
        logger.info(
            "User has configured app Proton version (CompatToolMapping): %s",
            tool_name
        )
        potential_names.append(tool_name)

    # Steam Deck compatibility profile has the 2nd highest priority
    if app_section and is_steam_deck():
        logger.info(
            "We're on a Steam Deck, checking if compatibility profile is "
            "available for the app"
        )
        recommended_runtime = (
            app_section["appinfo"]
            .get("common", {})
            .get("steam_deck_compatibility", {})
            .get("configuration", {})
            .get("recommended_runtime", None)
        )

        if recommended_runtime not in (None, "native"):
            logger.info(
                "App has Steam Deck compatibility profile with Proton "
                "version: %s",
                recommended_runtime
            )
            potential_names.append(recommended_runtime)

    # Game specific default compatibility tool mapping in Steam Play 2.0
    # manifest has the 3rd highest priority
    if manifest_app_compat_section:
        if "tool" in manifest_app_compat_section:
            tool = manifest_app_compat_section["tool"]
            logger.info(
                "App has default compatibility tool mapping in the Steam Play "
                "manifest: %s",
                tool
            )
            potential_names.append(tool)

    # Global user settings have the 4th highest priority
    if compat_tool_mapping.get("0", {}).get("name"):
        tool_name = compat_tool_mapping["0"]["name"]
        logger.info(
            "User has configured default Proton version (CompatToolMapping): "
            "%s",
            tool_name
        )
        potential_names.append(tool_name)

    # Legacy user settings (ToolMapping) have the 5th highest priority
    if tool_mapping.get(str(appid), {}).get("name", {}):
        tool_name = tool_mapping[str(appid)]["name"]
        logger.info(
            "User has configured app Proton version (ToolMapping): %s",
            tool_name
        )
        potential_names.append(tool_name)

    if tool_mapping.get("0", {}).get("name", {}):
        tool_name = tool_mapping["0"]["name"]
        logger.info(
            "User has configured default Proton version (ToolMapping): %s",
            tool_name
        )
        potential_names.append(tool_name)

    # Get the first name that was valid, or use stable Proton as fallback
    try:
        compat_tool_name = next(name for name in potential_names if name)
    except StopIteration:
        logger.info(
            "No compatibility tool found by reading Steam configuration. "
            "Using stable version of Proton as fallback."
        )
        compat_tool_name = "proton-stable"

    # We've got a compatibility tool name,
    # now there are two possible ways to find the installation
    # 1. It's a custom compatibility tool, and we simply need to find
    #    a SteamApp by its internal name
    # 2. It's a production Proton installation, in which case we need
    #    to parse a binary configuration file to find the App ID

    # Let's try option 1 first
    try:
        app = next(
            app for app in steam_apps if app.name == compat_tool_name
        )
        logger.info(
            "Found active custom compatibility tool: %s", app.name
        )
        return app
    except StopIteration:
        pass

    # Try option 2:
    # Find the corresponding App ID from <steam_path>/appcache/appinfo.vdf
    tool_app = _get_tool_app(
        compat_tool_name=compat_tool_name,
        steam_apps=steam_apps,
        steam_play_manifest=steam_play_manifest
    )

    if tool_app:
        logger.info(
            "Found active compatibility tool: %s", tool_app.name
        )
        return tool_app

    logger.error("Could not find configured Proton installation!")


def find_appid_proton_prefix(appid, steam_lib_paths):
    """
    Find the Proton prefix for the app by its App ID

    Proton prefix and the game installation itself can exist on different
    Steam libraries, making a search necessary
    """
    def get_prefix_modify_time(prefix_path):
        """
        Get the prefix modification time for sorting purposes.
        The newest modification time corresponds to the most recently
        used Proton prefix
        """
        try:
            # 'pfx.lock' is modified on game launch
            return (prefix_path.parent / "pfx.lock").stat().st_mtime
        except FileNotFoundError:
            return 0

    candidates = []

    for path in steam_lib_paths:
        steamapps_dirs = _get_steamapps_subdirs(path)
        for steamapps_path in steamapps_dirs:
            prefix_path = steamapps_path / "compatdata" / str(appid) / "pfx"
            if prefix_path.is_dir():
                candidates.append(prefix_path)

    logger.debug(
        "Found compatdata directories for app %s: %s", appid, candidates
    )

    if len(candidates) > 1:
        # If we have more than one possible prefix path, use the one
        # with the most recent modification date
        logger.info(
            "Multiple compatdata directories found for app %s", appid
        )
        candidates.sort(key=get_prefix_modify_time)
        candidates.reverse()

    if candidates:
        return candidates[0]

    return None


def find_proton_app(steam_path, steam_apps, appid=None):
    """
    Find the Proton app, using either $PROTON_VERSION or the one
    currently configured in Steam

    If 'appid' is provided, use it to find the app-specific Proton installation
    if one is configured
    """
    if os.environ.get("PROTON_VERSION"):
        proton_version = os.environ.get("PROTON_VERSION")
        try:
            proton_app = next(
                app for app in steam_apps if app.name == proton_version)
            logger.info(
                "Found requested Proton version: %s", proton_app.name
            )
            return proton_app
        except StopIteration:
            logger.error(
                "$PROTON_VERSION was set but matching Proton installation "
                "could not be found."
            )
            return None

    tool_app = find_steam_compat_tool_app(
        steam_path=steam_path, steam_apps=steam_apps, appid=appid)

    if not tool_app:
        logger.error(
            "Active Proton installation could not be found automatically."
        )
        return None

    # Check that it's actually a Proton app; Protontricks doesn't handle
    # other compatibility tools.
    if not tool_app.is_proton:
        logger.error(
            "Active compatibility tool was found, but it's not a Proton "
            "installation supported by Protontricks."
        )
        return None

    logger.info("Active compatibility tool is a Proton installation")

    return tool_app


def get_steam_lib_paths(steam_path):
    """
    Return a list of any Steam directories including any user-added
    Steam library folders
    """
    def parse_library_folders(data):
        """
        Parse the Steam library folders in the VDF file using the given data
        """
        vdf_data = lower_dict(vdf.loads(data))
        # Library folders have integer field names in ascending order
        library_entries = [
            value for key, value in vdf_data["libraryfolders"].items()
            if key.isdigit()
        ]
        library_folders = []

        for value in library_entries:
            if isinstance(value, dict):
                # Library data is stored in a dict in newer Steam releases
                path = Path(value["path"])
            else:
                # Older releases just store the library path as a string
                # and nothing else
                path = Path(value)

            xdg_steam_path = Path.home() / ".local/share/Steam"
            flatpak_steam_path = \
                Path.home() / ".var/app/com.valvesoftware.Steam/data/Steam"

            is_library_folder_xdg_steam = str(path) == str(xdg_steam_path)
            is_flatpak_steam = str(steam_path) == str(flatpak_steam_path)

            # Adjust the path if the library folder is "~/.local/share/Steam"
            # and we're looking for library folders in
            # "~/.var/app/com.valvesoftware.Steam"
            # This is because ~/.local/share/Steam inside a Steam Flatpak
            # sandbox corresponds to a different location, and we need to
            # adjust for that.
            if is_library_folder_xdg_steam and is_flatpak_steam:
                path = flatpak_steam_path

            logger.debug(
                "Found Steam library folder %s. Is Flatpak path: %s, "
                "Is XDG Steam path: %s.",
                path, is_flatpak_steam, is_library_folder_xdg_steam
            )

            library_folders.append(path)

        logger.info(
            "Found %d Steam library folders", len(library_folders)
        )
        return library_folders

    # Try finding Steam library folders using libraryfolders.vdf in Steam root
    try:
        steamapps_path = _get_steamapps_subdirs(steam_path)[0]
        folders_vdf_path = steamapps_path / "libraryfolders.vdf"
    except IndexError as exc:
        raise ValueError(
            "Steam installation directory does not contain a 'steamapps' "
            "directory. Installation may be incomplete or broken."
        ) from exc

    try:
        library_folders = parse_library_folders(folders_vdf_path.read_text())
    except OSError:
        # libraryfolders.vdf doesn't exist; maybe no Steam library folders
        # are set?
        library_folders = []
    except SyntaxError as exc:
        raise ValueError(
            f"Library folder configuration file {folders_vdf_path} is corrupted"
        ) from exc

    paths = [steam_path] + library_folders

    # Get rid of duplicate paths by fully resolving them and turning them into
    # a set and back
    paths = [path.resolve() for path in paths]
    paths = list(set(paths))

    return paths


def get_compat_tool_dirs(steam_root):
    """
    Return a list of compatibility tool directories in order from
    directories with lowest precedence
    """
    # The path list is ordered by priority, starting from Proton apps
    # with the lowest precedence ('/usr/share/steam/compatibilitytools.d')
    paths = [
        Path("/usr/share/steam/compatibilitytools.d"),
        Path("/usr/local/share/steam/compatibilitytools.d"),
    ]
    extra_ct_paths_env = os.getenv("STEAM_EXTRA_COMPAT_TOOLS_PATHS")
    if extra_ct_paths_env:
        logger.debug(
            "Including extra compat tool paths provided via env var: %s",
            extra_ct_paths_env
        )
        paths += [Path(path) for path in extra_ct_paths_env.split(os.pathsep)]
    paths += [steam_root / "compatibilitytools.d"]

    return paths


def get_custom_compat_tool_installations_in_dir(compat_tool_dir):
    """
    Return a list of custom compatibility tools in the given directory
    as a list of SteamApp objects
    """
    if not compat_tool_dir.is_dir():
        return []

    comptool_files = list(compat_tool_dir.glob("*/compatibilitytool.vdf"))
    comptool_files += list(compat_tool_dir.glob("compatibilitytool.vdf"))

    custom_tool_apps = []

    for vdf_path in comptool_files:
        content = vdf_path.read_text()

        logger.debug(
            "Parsing custom compatibility tool manifest at %s", vdf_path
        )

        try:
            vdf_data = vdf.loads(content)
        except SyntaxError:
            logger.warning(
                "Compatibility tool declaration at %s is corrupted. You may "
                "need to reinstall the application.",
                vdf_path
            )
            continue

        if not vdf_data:
            logger.warning(
                "Compatibility tool declaration at %s is empty. You may need "
                "to reinstall the application.",
                vdf_path
            )
            continue

        # Traverse to 'compatibilitytools/compat_tools' in a case-insensitive
        # way. This is done because we can't turn all keys recursively to
        # lowercase from the get-go; the app name is stored as a key.
        try:
            compat_tools = {k.lower(): v for k, v in vdf_data.items()}
            compat_tools = compat_tools["compatibilitytools"]
            compat_tools = {
                k.lower(): v for k, v in compat_tools.items()
            }
            compat_tools = compat_tools["compat_tools"]
            internal_name = list(compat_tools.keys())[0]
            tool_info = compat_tools[internal_name]

            # We can now convert the remainder into lowercase
            tool_info = lower_dict(tool_info)

            install_path_name = tool_info["install_path"]
            from_oslist = tool_info["from_oslist"]
            to_oslist = tool_info["to_oslist"]
        except KeyError:
            logger.warning(
                "Compatibility tool declaration at %s is incomplete. You may "
                "need to reinstall the application",
                vdf_path
            )
            continue

        if from_oslist != "windows" or to_oslist != "linux":
            continue

        # Installation path can be relative if the VDF was in
        # 'compatibilitytools.d/'
        # or '.' if the VDF was in 'compatibilitytools.d/TOOL_NAME'
        if install_path_name == ".":
            install_path = vdf_path.parent
        else:
            install_path = compat_tool_dir / install_path_name

        # Check if the app requires another app. This is the case with
        # newer versions of Proton, which use Steam Runtimes installed as
        # normal Steam apps
        try:
            required_tool_appid = _get_required_tool_appid(install_path)
        except (ValueError, SyntaxError):
            logger.warning(
                "Tool manifest for %s is empty or corrupted. You may need to "
                "reinstall the application.",
                install_path.name
            )
            continue

        logger.debug(
            "Found custom compatibility tool %s at %s",
            internal_name, install_path
        )

        custom_tool_apps.append(
            SteamApp(
                name=internal_name, install_path=install_path,
                required_tool_appid=required_tool_appid
            )
        )

    return custom_tool_apps


def get_custom_compat_tool_installations(steam_root):
    """
    Get a list of all custom compatibility tools as a list of SteamApp objects
    """
    custom_tool_apps = {}
    for dir_ in get_compat_tool_dirs(steam_root=steam_root):
        for tool_app in get_custom_compat_tool_installations_in_dir(dir_):
            # If another tool app exists with the same name, it will
            # be replaced with an installation that has higher precedence
            # here
            custom_tool_apps[tool_app.name] = tool_app

    # Return the list of tool apps as a list
    custom_tool_apps = list(custom_tool_apps.values())

    return custom_tool_apps


def find_current_steamid3(steam_path):
    """
    Find the SteamID3 of the currently logged in Steam user
    """
    def to_steamid3(steamid64):
        """Convert a SteamID64 into the SteamID3 format"""
        return int(steamid64) & 0xffffffff

    loginusers_path = steam_path / "config" / "loginusers.vdf"
    try:
        content = loginusers_path.read_text()
        vdf_data = lower_dict(vdf.loads(content))
    except IOError:
        logger.warning(
            "Couldn't determine the currently logged-in Steam user. Custom "
            "shortcuts won't be detected."
        )
        return None

    user_datas = [
        (user_id, lower_dict(user_data))
        for user_id, user_data in vdf_data["users"].items()
    ]
    users = [
        {
            "steamid3": to_steamid3(user_id),
            "account_name": user_data["accountname"],
            "timestamp": user_data.get("timestamp", 0)
        }
        for user_id, user_data in user_datas
    ]

    logger.debug("Found Steam user entries: %s", users)

    # Return the user with the highest timestamp, as that's likely to be the
    # currently logged-in user
    if users:
        user = max(users, key=lambda u: u["timestamp"])
        logger.info(
            "Currently logged-in Steam user: %s", user["account_name"]
        )
        return user["steamid3"]

    return None


def get_appid_from_shortcut(target, name):
    """
    Get the identifier used for the Proton prefix from a shortcut's
    target and name
    """
    # First, calculate the screenshot ID Steam uses for shortcuts
    data = b"".join([
        target.encode("utf-8"),
        name.encode("utf-8")
    ])
    result = zlib.crc32(data) & 0xffffffff
    result = result | 0x80000000
    result = (result << 32) | 0x02000000

    # Derive the prefix ID from the screenshot ID
    return result >> 32


def get_custom_windows_shortcuts(steam_path, steam_lib_paths):
    """
    Get a list of custom shortcuts for Windows applications as a list
    of SteamApp objects
    """
    # Get the Steam ID3 for the currently logged-in user
    steamid3 = find_current_steamid3(steam_path)

    shortcuts_path = \
        steam_path / "userdata" / str(steamid3) / "config" / "shortcuts.vdf"

    try:
        content = shortcuts_path.read_bytes()
        # Tolerate VDF files that have extra data after the binary VDF section.
        # Steam itself can supposedly create such files in some situations.
        vdf_data = lower_dict(
            vdf.binary_loads(content, raise_on_remaining=False)
        )
    except IOError:
        logger.info(
            "Couldn't find custom shortcuts. Maybe none have been created yet?"
        )
        return []

    steam_apps = []

    for shortcut_id, shortcut_data in vdf_data["shortcuts"].items():
        # The "exe" field can also be "Exe". Account for this by making
        # all field names lowercase
        shortcut_data = lower_dict(shortcut_data)
        shortcut_id = int(shortcut_id)

        if "appid" in shortcut_data:
            try:
                appid = shortcut_data["appid"] & 0xffffffff
            except TypeError:
                logger.info(
                    "Skipping unrecognized non-Steam shortcut with app ID "
                    "'%s'",
                    shortcut_data["appid"]
                )
                continue
        else:
            appid = get_appid_from_shortcut(
                target=shortcut_data["exe"], name=shortcut_data["appname"]
            )

        prefix_path = find_appid_proton_prefix(
            appid=appid, steam_lib_paths=steam_lib_paths
        )
        if not prefix_path:
            logger.info(
                "Shortcut %s (%s) does not have a prefix. "
                "It's either not a Proton app or it hasn't been launched yet.",
                shortcut_data["appname"], appid
            )
            continue

        install_path = Path(shortcut_data["startdir"].strip('"'))

        try:
            # Check that we have permission to access the installation
            # directory, whether it exists or not. Any attempts to check
            # any files will fail later if this isn't done.
            install_path.is_dir()

            if not prefix_path.is_dir():
                continue
        except PermissionError as exc:
            logger.warning(
                "Skipping shortcut %s due to insufficient permissions. "
                "Error: %s",
                shortcut_data["appname"], str(exc)
            )
            continue

        icon_path = None

        if shortcut_data.get("icon", None):
            icon_path = Path(shortcut_data["icon"])

        logger.debug(
            "Creating SteamApp from non-Steam shortcut in %s", install_path
        )

        steam_apps.append(
            SteamApp(
                appid=appid,
                name=f"Non-Steam shortcut: {shortcut_data['appname']}",
                prefix_path=prefix_path,
                install_path=install_path,
                icon_path=icon_path

            )
        )

    logger.info(
        "Found %d Steam shortcuts running using Steam compatibility tools",
        len(steam_apps)
    )

    return steam_apps


def _link_tool_apps(steam_apps):
    """
    Check which Steam apps require other Steam apps and add the corresponding
    references
    """
    appid2steam_app = {steam_app.appid: steam_app for steam_app in steam_apps}

    for steam_app in steam_apps:
        if steam_app.required_tool_appid:
            steam_app.required_tool_app = \
                appid2steam_app.get(steam_app.required_tool_appid)


def get_steam_apps(steam_root, steam_path, steam_lib_paths):
    """
    Find all the installed Steam apps and return them as a list of SteamApp
    objects
    """
    steam_apps = []

    for path in steam_lib_paths:
        try:
            if not path.is_dir():
                logger.warning(
                    "Steam library folder %s not found. Protontricks "
                    "might not have access to the directory.",
                    str(path)
                )
                continue
        except PermissionError:
            logger.warning(
                "Skipping library folder %s due to insufficient permissions",
                str(path)
            )
            continue

        appmanifest_paths = []
        steamapps_dirs = _get_steamapps_subdirs(path)

        try:
            appmanifest_paths = steamapps_dirs[0].glob("appmanifest_*.acf")
        except IndexError:
            logger.warning(
                "No 'steamapps' directory was found at %s", str(path)
            )

        if len(steamapps_dirs) > 1:
            # Log a warning if multiple 'steamapps' directories with different
            # cases exist, as both Protontricks and Steam client have problems
            # dealing with them (see issue #51)
            logger.warning(
                "Multiple 'steamapps' directories were found "
                "at %s. Only one directory should exist to prevent issues "
                "with app and Proton discovery.",
                str(path)
            )

        for manifest_path in appmanifest_paths:
            try:
                steam_app = SteamApp.from_appmanifest(
                    manifest_path, steam_lib_paths=steam_lib_paths,
                    steam_path=steam_path
                )
            except PermissionError as exc:
                logger.warning(
                    "Could not load manifest %s due to insufficient "
                    "permissions. Error: %s",
                    str(manifest_path), str(exc)
                )
                steam_app = None

            if steam_app:
                steam_apps.append(steam_app)

    # Get the custom compatibility tools and non-Steam shortcuts as well
    steam_apps += get_custom_compat_tool_installations(steam_root=steam_root)
    steam_apps += get_custom_windows_shortcuts(
        steam_path=steam_path, steam_lib_paths=steam_lib_paths
    )

    # Exclude games that haven't been launched yet
    steam_apps = [
        app for app in steam_apps
        if app.prefix_path_exists or app.is_proton or app.is_tool
    ]

    # Populate the `SteamApp.required_tool_app` parameter for Steam apps
    # which rely on other Steam apps
    _link_tool_apps(steam_apps)

    # Sort the apps by their names
    steam_apps.sort(key=lambda app: app.name)

    return steam_apps



import stat
import textwrap
from pathlib import Path

import pytest

from protontricks.util import (create_wine_bin_dir, is_steam_deck, lower_dict,
                               run_command)


def get_files_in_dir(d):
    return {binary.name for binary in d.iterdir()}


class TestCreateWineBinDir:
    def test_wine_bin_dir_updated(self, home_dir, default_proton):
        """
        Test that the directory containing the helper scripts is kept
        up-to-date with the Proton installation's binaries
        """
        create_wine_bin_dir(default_proton)

        # Check that the Wine binaries exist
        files = get_files_in_dir(
            home_dir / ".cache" / "protontricks" / "proton" / "Proton 4.20"
            / "bin"
        )
        assert set([
            "wine", "wineserver", "wineserver-keepalive", "bwrap-launcher",
            "wineserver-keepalive.bat"
        ]) == files

        # Create a new binary for the Proton installation and delete another
        # one
        proton_bin_path = Path(default_proton.install_path) / "dist" / "bin"

        (proton_bin_path / "winedine").touch()
        (proton_bin_path / "wineserver").unlink()

        # The old scripts will be deleted and regenerated now that the Proton
        # installation's contents changed
        create_wine_bin_dir(default_proton)

        files = get_files_in_dir(
            home_dir / ".cache" / "protontricks" / "proton" / "Proton 4.20"
            / "bin"
        )
        # Scripts are regenerated
        assert set([
            "wine", "winedine", "wineserver-keepalive", "bwrap-launcher",
            "wineserver-keepalive.bat"
        ]) == files


class TestRunCommand:
    def test_user_environment_variables_used(
            self, default_proton, steam_runtime_dir, steam_app_factory,
            home_dir, command_mock, monkeypatch):
        """
        Test that user-provided environment variables are used even when
        Steam Runtime is enabled
        """
        steam_app = steam_app_factory(name="Fake game", appid=10)

        run_command(
            winetricks_path=Path("/usr/bin/winetricks"),
            proton_app=default_proton,
            steam_app=steam_app,
            command=["echo", "nothing"],
            use_steam_runtime=True,
            legacy_steam_runtime_path=steam_runtime_dir / "steam-runtime"
        )

        # Proxy scripts are used if no environment variables are set by the
        # user
        wine_bin_dir = (
            home_dir / ".cache" / "protontricks" / "proton" / "Proton 4.20"
            / "bin"
        )

        command = command_mock.commands[-1]
        assert command.args == ["echo", "nothing"]
        assert command.env["WINE"] == str(wine_bin_dir / "wine")
        assert command.env["WINELOADER"] == str(wine_bin_dir / "wine")
        assert command.env["WINESERVER"] == str(wine_bin_dir / "wineserver")

        monkeypatch.setenv("WINE", "/fake/wine")
        monkeypatch.setenv("WINESERVER", "/fake/wineserver")

        run_command(
            winetricks_path=Path("/usr/bin/winetricks"),
            proton_app=default_proton,
            steam_app=steam_app,
            command=["echo", "nothing"],
            use_steam_runtime=True,
            legacy_steam_runtime_path=steam_runtime_dir / "steam-runtime"
        )

        # User provided Wine paths are used even when Steam Runtime is enabled
        command = command_mock.commands[-1]
        assert command.args == ["echo", "nothing"]
        assert command.env["WINE"] == "/fake/wine"
        assert command.env["WINELOADER"] == "/fake/wine"
        assert command.env["WINESERVER"] == "/fake/wineserver"

    @pytest.mark.usefixtures("command_mock")
    def test_unknown_steam_runtime_detected(
            self, home_dir, proton_factory, runtime_app_factory,
            steam_app_factory, caplog):
        """
        Test that Protontricks will log a warning if it encounters a Steam
        Runtime it does not recognize
        """
        steam_runtime_medic = runtime_app_factory(
            name="Steam Linux Runtime - Medic",
            appid=14242420,
            runtime_dir_name="medic"
        )
        proton_app = proton_factory(
            name="Proton 5.20", appid=100, compat_tool_name="proton_520",
            is_default_proton=True, required_tool_app=steam_runtime_medic
        )
        steam_app = steam_app_factory(name="Fake game", appid=10)

        run_command(
            winetricks_path=Path("/usr/bin/winetricks"),
            proton_app=proton_app,
            steam_app=steam_app,
            command=["echo", "nothing"],
            shell=True,
            use_steam_runtime=True
        )

        # Warning will be logged since Protontricks does not recognize
        # Steam Runtime Medic and can't ensure it's being configured correctly
        warning = next(
            record for record in caplog.records
            if record.levelname == "WARNING"
            and "not recognized" in record.getMessage()
        )
        assert warning.getMessage() == \
            "Current Steam Runtime not recognized by Protontricks."

    @pytest.mark.usefixtures("steam_deck")
    def test_locale_fixed_on_steam_deck(
            self, proton_factory, default_proton, steam_app_factory, home_dir,
            command_mock, caplog):
        """
        Test that Protontricks will fix locale settings if nonexistent locale
        settings are detected and Steam Deck is used to run Protontricks
        """
        # Create binary to fake the 'locale' executable
        locale_script_path = home_dir / ".local" / "bin" / "locale"
        locale_script_path.write_text("""#!/bin/sh
            if [ "$1" = "-a" ]; then
                echo 'C'
                echo 'C.UTF-8'
                echo 'en_US'
                echo 'en_US.utf8'
            else
                echo 'LANG=fi_FI.UTF-8'
                echo 'LC_CTYPE=en_US.utf8'
                echo 'LC_TIME=en_US.UTF-8'
                echo 'LC_NUMERIC=D'
            fi
        """)
        locale_script_path.chmod(
            locale_script_path.stat().st_mode | stat.S_IEXEC
        )

        steam_app = steam_app_factory(name="Fake game", appid=10)
        run_command(
            winetricks_path=Path("/usr/bin/winetricks"),
            proton_app=default_proton,
            steam_app=steam_app,
            command=["/bin/env"],
            env={
                # Use same environment variables as in the mocked 'locale'
                # script
                "LANG": "fi_FI.UTF-8",
                "LC_CTYPE": "en_US.utf8",
                "LC_TIME": "en_US.UTF-8",
                "LC_NUMERIC": "D"
            }
        )

        # Warning will be logged to indicate 'LANG' was changed
        warning = next(
            record for record in caplog.records
            if record.levelname == "WARNING"
            and "locale has been reset" in record.getMessage()
        )
        assert warning.getMessage().endswith(
            "for the following categories: LANG, LC_NUMERIC"
        )

        # Ensure the incorrect locale settings were changed for the command
        command = command_mock.commands[-1]
        assert command.env["LANG"] == "en_US.UTF-8"
        # LC_CTYPE was not changed as 'en_US.UTF-8' and 'en_US.utf8'
        # are identical after normalization.
        assert command.env["LC_CTYPE"] == "en_US.utf8"
        assert command.env["LC_TIME"] == "en_US.UTF-8"
        assert command.env["LC_NUMERIC"] == "en_US.UTF-8"

    def test_bwrap_launcher_crash_detected(
            self, default_new_proton, steam_app_factory, command_mock):
        """
        Test that Protontricks will raise an exception if `bwrap-launcher`
        crashes unexpectedly
        """
        steam_app = steam_app_factory(name="Fake game", appid=10)

        # Mock a crashing 'bwrap-launcher'
        command_mock.launcher_working = False

        with pytest.raises(RuntimeError) as exc:
            run_command(
                winetricks_path=Path("/usr/bin/winetricks"),
                proton_app=default_new_proton,
                steam_app=steam_app,
                command=["echo", "nothing"],
                shell=True,
                use_steam_runtime=True
            )

        assert str(exc.value) == "bwrap launcher crashed, returncode: 1"


class TestLowerDict:
    def test_lower_nested_dict(self):
        """
        Turn all keys in a nested dictionary to lowercase using `lower_dict`
        """
        before = {
            "AppState": {
                "Name": "Blah",
                "appid": 123450,
                "userconfig": {
                    "Language": "English"
                }
            }
        }

        after = {
            "appstate": {
                "name": "Blah",
                "appid": 123450,
                "userconfig": {
                    "language": "English"
                }
            }
        }

        assert lower_dict(before) == after


class TestIsSteamDeck:
    def test_not_steam_deck(self):
        """
        Test that non-Steam Deck environment is detected correctly
        """
        assert not is_steam_deck()

    @pytest.mark.usefixtures("steam_deck")
    def test_is_steam_deck(self):
        """
        Test that Steam Deck environment is detected correctly
        """
        assert is_steam_deck()



import argparse
import atexit
import functools
import logging
import os
import sys
import tempfile
import traceback
from pathlib import Path

from ..gui import show_text_dialog


def _get_log_file_path():
    """
    Get the log file path to use for this Protontricks process.
    """
    temp_dir = tempfile.gettempdir()

    pid = os.getpid()
    return Path(temp_dir) / f"protontricks{pid}.log"


def _delete_log_file():
    """
    Delete the log file if one exists.

    This is usually executed before shutdown by registering this function
    using `atexit`
    """
    try:
        _get_log_file_path().unlink()
    except FileNotFoundError:
        pass


def enable_logging(level=0, record_to_file=True):
    """
    Enables logging.

    :param int level: Level of logging. 0 = WARNING, 1 = INFO, 2 = DEBUG.
    :param bool record_to_file: Whether to log the generated log messages
                                to a temporary file.
                                This is used for the error dialog containing
                                log records.
    """
    if level >= 2:
        level = logging.DEBUG
        label = "DEBUG"
    elif level >= 1:
        level = logging.INFO
        label = "INFO"
    else:
        level = logging.WARNING
        label = "WARNING"

    # 'PROTONTRICKS_LOG_LEVEL' env var allows separate Bash scripts
    # to detect when logging is enabled.
    os.environ["PROTONTRICKS_LOG_LEVEL"] = label

    logger = logging.getLogger("protontricks")

    stream_handler_added = any(
        filter(
            lambda hndl: hndl.name == "protontricks-stream", logger.handlers
        )
    )

    if not stream_handler_added:
        # Logs printed to stderr will follow the log level
        stream_handler = logging.StreamHandler()
        stream_handler.name = "protontricks-stream"
        stream_handler.setLevel(level)
        stream_handler.setFormatter(
            logging.Formatter("%(name)s (%(levelname)s): %(message)s")
        )

        logger.setLevel(logging.DEBUG)
        logger.addHandler(stream_handler)

        logger.debug("Stream log handler added")

    if not record_to_file:
        return

    file_handler_added = any(
        filter(lambda hndl: hndl.name == "protontricks-file", logger.handlers)
    )

    if not file_handler_added:
        # Record log files to temporary file. This means log messages can be
        # printed at the end of the session in an error dialog.
        # INFO and WARNING log messages are written into this file whether
        # `--verbose` is enabled or not.
        log_file_path = _get_log_file_path()
        try:
            log_file_path.unlink()
        except FileNotFoundError:
            pass

        file_handler = logging.FileHandler(str(_get_log_file_path()))
        file_handler.name = "protontricks-file"
        file_handler.setLevel(logging.INFO)
        logger.addHandler(file_handler)

        # Ensure the log file is removed before the process exits
        atexit.register(_delete_log_file)

        logger.debug("File log handler added")


def exit_with_error(error, desktop=False):
    """
    Exit with an error, either by printing the error to stderr or displaying
    an error dialog.

    :param bool desktop: If enabled, display an error dialog containing
                         the error itself and additional log messages.
    """
    if not desktop:
        print(error)
        sys.exit(1)

    try:
        log_messages = _get_log_file_path().read_text()
    except FileNotFoundError:
        log_messages = "!! LOG FILE NOT FOUND !!"

    # Display an error dialog containing the message
    message = "".join([
        "Protontricks was closed due to the following error:\n\n",
        f"{error}\n\n",
        "=============\n\n",
        "Please include this entire error message when making a bug report.\n",
        "Log messages:\n\n",
        f"{log_messages}"
    ])

    show_text_dialog(
        title="Protontricks",
        text=message,
        window_icon=error
    )
    sys.exit(1)


def cli_error_handler(cli_func):
    """
    Decorator for CLI entry points.

    If an unhandled exception is raised and Protontricks was launched from
    desktop, display an error dialog containing the stack trace instead
    of printing to stderr.
    """
    @functools.wraps(cli_func)
    def wrapper(self, *args, **kwargs):
        try:
            wrapper.no_term = False
            return cli_func(self, *args, **kwargs)
        except Exception:  # pylint: disable=broad-except
            if not wrapper.no_term:
                # If we weren't launched from desktop, handle it normally
                raise

            traceback_ = traceback.format_exc()
            exit_with_error(traceback_, desktop=True)

    return wrapper


class CustomArgumentParser(argparse.ArgumentParser):
    """
    Custom argument parser that prints the full help message
    when incorrect parameters are provided
    """
    def error(self, message):
        self.print_help(sys.stderr)
        args = {'prog': self.prog, 'message': message}
        self.exit(2, '%(prog)s: error: %(message)s\n' % args)



import contextlib
import shutil
from subprocess import CalledProcessError

import pytest
from conftest import MockResult
from PIL import Image

from protontricks.gui import (prompt_filesystem_access,
                              select_steam_app_with_gui,
                              select_steam_installation)


@pytest.fixture(scope="function")
def broken_zenity(gui_provider, monkeypatch):
    """
    Mock a broken Zenity executable that prints an error as described in
    the following GitHub issue:
    https://github.com/Matoking/protontricks/issues/20
    """
    def mock_subprocess_run(args, **kwargs):
        gui_provider.args = args

        raise CalledProcessError(
            returncode=-6,
            cmd=args,
            output=gui_provider.mock_stdout,
            stderr=b"free(): double free detected in tcache 2\n"
        )

    monkeypatch.setattr(
        "protontricks.gui.run",
        mock_subprocess_run
    )

    yield gui_provider


@pytest.fixture(scope="function")
def locale_error_zenity(gui_provider, monkeypatch):
    """
    Mock a Zenity executable returning a 255 error due to a locale issue
    on first run and working normally on second run
    """
    def mock_subprocess_run(args, **kwargs):
        if not gui_provider.args:
            gui_provider.args = args
            raise CalledProcessError(
                returncode=255,
                cmd=args,
                output="",
                stderr=(
                    b"This option is not available. "
                    b"Please see --help for all possible usages."
                )
            )

        return MockResult(stdout=gui_provider.mock_stdout.encode("utf-8"))

    monkeypatch.setattr(
        "protontricks.gui.run",
        mock_subprocess_run
    )
    monkeypatch.setenv("PROTONTRICKS_GUI", "zenity")

    yield gui_provider


class TestSelectApp:
    def test_select_game(self, gui_provider, steam_app_factory, steam_dir):
        """
        Select a game using the GUI
        """
        steam_apps = [
            steam_app_factory(name="Fake game 1", appid=10),
            steam_app_factory(name="Fake game 2", appid=20)
        ]

        # Fake user selecting 'Fake game 2'
        gui_provider.mock_stdout = "Fake game 2: 20"
        steam_app = select_steam_app_with_gui(
            steam_apps=steam_apps, steam_path=steam_dir
        )

        assert steam_app == steam_apps[1]

        input_ = gui_provider.kwargs["input"]

        # Check that choices were displayed
        assert b"Fake game 1: 10\n" in input_
        assert b"Fake game 2: 20" in input_

    def test_select_game_icons(
            self, gui_provider, steam_app_factory, steam_dir):
        """
        Select a game using the GUI. Ensure that icons are used in the dialog
        whenever available.
        """
        steam_apps = [
            steam_app_factory(name="Fake game 1", appid=10),
            steam_app_factory(name="Fake game 2", appid=20),
            steam_app_factory(name="Fake game 3", appid=30),
        ]

        # Create icons for game 1 and 3
        for appid in (10, 30):
            Image.new("RGB", (32, 32)).save(
                steam_dir / "appcache" / "librarycache" / f"{appid}_icon.jpg")

        gui_provider.mock_stdout = "Fake game 2: 20"
        select_steam_app_with_gui(steam_apps=steam_apps, steam_path=steam_dir)

        input_ = gui_provider.kwargs["input"]

        assert b"librarycache/10_icon.jpg\nFake game 1" in input_
        assert b"icon_placeholder.png\nFake game 2" in input_
        assert b"librarycache/30_icon.jpg\nFake game 3" in input_

    def test_select_game_icons_ensure_resize(
            self, gui_provider, steam_app_factory, steam_dir, home_dir):
        """
        Select a game using the GUI. Ensure custom icons with sizes other than
        32x32 are resized.
        """
        steam_apps = [
            steam_app_factory(name="Fake game 1", appid=10)
        ]

        Image.new("RGB", (64, 64)).save(
            steam_dir / "appcache" / "librarycache" / "10_icon.jpg"
        )

        gui_provider.mock_stdout = "Fake game 1: 10"
        select_steam_app_with_gui(steam_apps=steam_apps, steam_path=steam_dir)

        # Resized icon should have been created with the correct size and used
        resized_icon_path = \
            home_dir / ".cache" / "protontricks" / "app_icons" / "10.jpg"
        assert resized_icon_path.is_file()
        with Image.open(resized_icon_path) as img:
            assert img.size == (32, 32)

        input_ = gui_provider.kwargs["input"]

        assert f"{resized_icon_path}\nFake game 1".encode("utf-8") in input_

        # Any existing icon should be overwritten if it already exists
        resized_icon_path.write_bytes(b"not valid")
        select_steam_app_with_gui(steam_apps=steam_apps, steam_path=steam_dir)

        with Image.open(resized_icon_path) as img:
            assert img.size == (32, 32)

    def test_select_game_unidentifiable_icon_skipped(
            self, gui_provider, steam_app_factory, steam_dir, home_dir, caplog):
        """
        Select a game using the GUI. Ensure a custom icon that's not
        identifiable by Pillow is skipped.
        """
        steam_apps = [
            steam_app_factory(name="Fake game 1", appid=10)
        ]

        icon_path = steam_dir / "appcache" / "librarycache" / "10_icon.jpg"
        icon_path.write_bytes(b"")

        gui_provider.mock_stdout = "Fake game 1: 10"
        selected_app = select_steam_app_with_gui(
            steam_apps=steam_apps, steam_path=steam_dir
        )

        # Warning about icon was logged, but the app was selected successfully
        record = caplog.records[-1]
        assert record.message.startswith(f"Could not resize {icon_path}")

        assert selected_app.appid == 10

    def test_select_game_no_choice(
            self, gui_provider, steam_app_factory, steam_dir):
        """
        Try choosing a game but make no choice
        """
        steam_apps = [steam_app_factory(name="Fake game 1", appid=10)]

        # Fake user doesn't select any game
        gui_provider.mock_stdout = ""

        with pytest.raises(SystemExit) as exc:
            select_steam_app_with_gui(
                steam_apps=steam_apps, steam_path=steam_dir
            )

        assert exc.value.code == 1

    def test_select_game_broken_zenity(
            self, broken_zenity, monkeypatch, steam_app_factory, steam_dir):
        """
        Try choosing a game with a broken Zenity executable that
        prints a specific error message that Protontricks knows how to ignore
        """
        monkeypatch.setenv("PROTONTRICKS_GUI", "zenity")

        steam_apps = [
            steam_app_factory(name="Fake game 1", appid=10),
            steam_app_factory(name="Fake game 2", appid=20)
        ]

        # Fake user selecting 'Fake game 2'
        broken_zenity.mock_stdout = "Fake game 2: 20"
        steam_app = select_steam_app_with_gui(
            steam_apps=steam_apps, steam_path=steam_dir)

        assert steam_app == steam_apps[1]

    def test_select_game_locale_error(
            self, locale_error_zenity, steam_app_factory, steam_dir, caplog):
        """
        Try choosing a game with an environment that can't handle non-ASCII
        characters
        """
        steam_apps = [
            steam_app_factory(name="Fäke game 1", appid=10),
            steam_app_factory(name="Fäke game 2", appid=20)
        ]

        # Fake user selecting 'Fäke game 2'. The non-ASCII character 'ä'
        # is stripped since Zenity wouldn't be able to display the character.
        locale_error_zenity.mock_stdout = "Fke game 2: 20"
        steam_app = select_steam_app_with_gui(
            steam_apps=steam_apps, steam_path=steam_dir
        )

        assert steam_app == steam_apps[1]
        assert (
            "Your system locale is incapable of displaying all characters"
            in caplog.records[-1].message
        )

    @pytest.mark.parametrize("gui_cmd", ["yad", "zenity"])
    def test_select_game_gui_provider_env(
            self, gui_provider, steam_app_factory, monkeypatch, gui_cmd,
            steam_dir):
        """
        Test that the correct GUI provider is selected based on the
        `PROTONTRICKS_GUI` environment variable
        """
        monkeypatch.setenv("PROTONTRICKS_GUI", gui_cmd)

        steam_apps = [
            steam_app_factory(name="Fake game 1", appid=10),
            steam_app_factory(name="Fake game 2", appid=20)
        ]

        gui_provider.mock_stdout = "Fake game 2: 20"
        select_steam_app_with_gui(
            steam_apps=steam_apps, steam_path=steam_dir
        )

        # The flags should differ slightly depending on which provider is in
        # use
        if gui_cmd == "yad":
            assert gui_provider.args[0] == "yad"
            assert gui_provider.args[2] == "--no-headers"
        elif gui_cmd == "zenity":
            assert gui_provider.args[0] == "zenity"
            assert gui_provider.args[2] == "--hide-header"


class TestSelectSteamInstallation:
    @pytest.mark.usefixtures("flatpak_sandbox")
    @pytest.mark.parametrize("gui_cmd", ["yad", "zenity"])
    def test_select_steam_gui_provider_env(
            self, gui_provider, monkeypatch, gui_cmd, steam_dir,
            flatpak_steam_dir):
        """
        Test that the correct GUI provider is selected based on the
        `PROTONTRICKS_GUI` environment variable
        """
        monkeypatch.setenv("PROTONTRICKS_GUI", gui_cmd)

        gui_provider.mock_stdout = "1: Flatpak - /foo/bar"
        select_steam_installation([
            (steam_dir, steam_dir),
            (flatpak_steam_dir, flatpak_steam_dir)
        ])

        # The flags should differ slightly depending on which provider is in
        # use
        if gui_cmd == "yad":
            assert gui_provider.args[0] == "yad"
            assert gui_provider.args[2] == "--no-headers"
        elif gui_cmd == "zenity":
            assert gui_provider.args[0] == "zenity"
            assert gui_provider.args[2] == "--hide-header"

    @pytest.mark.parametrize(
        "path,label",
        [
            (".steam", "Native"),
            (".local/share/Steam", "Native"),
            (".var/app/com.valvesoftware.Steam/.local/share/Steam", "Flatpak"),
            ("snap/steam/common/.local/share/Steam", "Snap")
        ]
    )
    def test_correct_labels_detected(
            self, gui_provider, steam_dir, home_dir, path, label):
        """
        Test that the Steam installation selection dialog uses the correct
        label for each Steam installation depending on its type
        """
        steam_new_dir = home_dir / path
        with contextlib.suppress(FileExistsError):
            # First test cases try copying against existing dirs, this can be
            # ignored
            shutil.copytree(steam_dir, steam_new_dir)

        select_steam_installation([
            (steam_new_dir, steam_new_dir),
            # Use an additional nonsense path; there need to be at least
            # two paths or user won't be prompted as there is no need
            ("/mock-steam", "/mock-steam")
        ])

        prompt_input = gui_provider.kwargs["input"].decode("utf-8")

        assert f"{label} - {steam_new_dir}" in prompt_input


@pytest.mark.usefixtures("flatpak_sandbox")
class TestPromptFilesystemAccess:
    def test_prompt_without_desktop(self, home_dir, caplog):
        """
        Test that calling 'prompt_filesystem_access' without showing the dialog
        only generates a warning
        """
        prompt_filesystem_access(
            [home_dir / "fake_path", "/mnt/fake_SSD", "/mnt/fake_SSD_2"],
            show_dialog=False
        )

        assert len(caplog.records) == 1

        record = caplog.records[0]

        assert record.levelname == "WARNING"
        assert "Protontricks does not appear to have access" in record.message

        assert "--filesystem=/mnt/fake_SSD" in record.message
        assert "--filesystem=/mnt/fake_SSD_2" in record.message
        assert str(home_dir / "fake_path") not in record.message

    def test_prompt_home_dir(self, home_dir, tmp_path, caplog):
        """
        Test that calling 'prompt_filesystem_access' with a path
        in the home directory will result in the command using a tilde slash
        as the shorthand instead
        """
        flatpak_info_path = tmp_path / "flatpak-info"

        flatpak_info_path.write_text(
            "[Application]\n"
            "name=fake.flatpak.Protontricks\n"
            "\n"
            "[Instance]\n"
            "flatpak-version=1.12.1\n"
            "\n"
            "[Context]\n"
            "filesystems=/mnt/SSD_A"
        )
        prompt_filesystem_access(
            [home_dir / "fake_path", "/mnt/SSD_A"],
            show_dialog=False
        )

        assert len(caplog.records) == 1

        record = caplog.records[0]

        assert record.levelname == "WARNING"
        assert "Protontricks does not appear to have access" in record.message

        assert "--filesystem='~/fake_path'" in record.message
        assert "/mnt/SSD_A" not in record.message

    def test_prompt_with_desktop_no_dialog(self, home_dir, gui_provider):
        """
        Test that calling 'prompt_filesystem_access' with 'show_dialog'
        displays a dialog
        """
        prompt_filesystem_access(
            [home_dir / "fake_path", "/mnt/fake_SSD", "/mnt/fake_SSD_2"],
            show_dialog=True
        )

        input_ = gui_provider.kwargs["input"].decode("utf-8")

        assert str(home_dir / "fake_path") not in input_
        assert "--filesystem=/mnt/fake_SSD" in input_
        assert "--filesystem=/mnt/fake_SSD_2" in input_

    def test_prompt_with_desktop_dialog(self, home_dir, gui_provider):
        """
        Test that calling 'prompt_filesystem_access' with 'show_dialog'
        displays a dialog
        """
        # Mock the user closing the dialog without ignoring the messages
        gui_provider.returncode = 1

        prompt_filesystem_access(
            [home_dir / "fake_path", "/mnt/fake_SSD", "/mnt/fake_SSD_2"],
            show_dialog=True
        )

        input_ = gui_provider.kwargs["input"].decode("utf-8")

        # Dialog was displayed
        assert "/mnt/fake_SSD" in input_
        assert "/mnt/fake_SSD_2" in input_

        # Mock the user selecting "Ignore, don't ask again"
        gui_provider.returncode = 0
        gui_provider.kwargs["input"] = None

        prompt_filesystem_access(
            [home_dir / "fake_path", "/mnt/fake_SSD", "/mnt/fake_SSD_2"],
            show_dialog=True
        )

        # Dialog is still displayed, but it won't be the next time
        input_ = gui_provider.kwargs["input"].decode("utf-8")
        assert "/mnt/fake_SSD" in input_
        assert "/mnt/fake_SSD_2" in input_

        gui_provider.kwargs["input"] = None

        prompt_filesystem_access(
            [home_dir / "fake_path", "/mnt/fake_SSD", "/mnt/fake_SSD_2"],
            show_dialog=True
        )

        # Dialog is not shown, since the user has opted to ignore the warning
        # for the current paths
        assert not gui_provider.kwargs["input"]

        # A new path makes the warning reappear
        prompt_filesystem_access(
            [
                home_dir / "fake_path",
                "/mnt/fake_SSD",
                "/mnt/fake_SSD_2",
                "/mnt/fake_SSD_3"
            ],
            show_dialog=True
        )

        input_ = gui_provider.kwargs["input"].decode("utf-8")
        assert "/mnt/fake_SSD " not in input_
        assert "/mnt/fake_SSD_2" not in input_
        assert "/mnt/fake_SSD_3" in input_
